@@ -23,10 +23,23 @@ open Clflags
 open Misc
 open Cmm
 
-type error = Assembler_error of string
+type error =
+  | Assembler_error of string
+  | Mismatched_for_pack of string option
+  | Asm_generation of string * Emitaux.error
 
 exception Error of error
 
+let cmm_invariants ppf fd_cmm =
+  let print_fundecl =
+    if !Clflags.dump_cmm then Printcmm.fundecl
+    else fun ppf fdecl -> Format.fprintf ppf "%s" fdecl.fun_name
+  in
+  if !Clflags.cmm_invariants && Cmm_invariants.run ppf fd_cmm then
+    Misc.fatal_errorf "Cmm invariants failed on following fundecl:@.%a@."
+      print_fundecl fd_cmm;
+  fd_cmm
+
 let liveness phrase = Liveness.fundecl phrase; phrase
 
 let dump_if ppf flag message phrase =
@@ -39,6 +52,44 @@ let pass_dump_linear_if ppf flag message phrase =
   if !flag then fprintf ppf "*** %s@.%a@." message Printlinear.fundecl phrase;
   phrase
 
+let start_from_emit = ref true
+
+let should_save_before_emit () =
+  should_save_ir_after Compiler_pass.Scheduling && (not !start_from_emit)
+
+let linear_unit_info =
+  { Linear_format.unit_name = "";
+    items = [];
+    for_pack = None;
+  }
+
+let reset () =
+  start_from_emit := false;
+  if should_save_before_emit () then begin
+    linear_unit_info.unit_name <- Compilenv.current_unit_name ();
+    linear_unit_info.items <- [];
+    linear_unit_info.for_pack <- !Clflags.for_package;
+  end
+
+let save_data dl =
+  if should_save_before_emit () then begin
+    linear_unit_info.items <- Linear_format.(Data dl) :: linear_unit_info.items
+  end;
+  dl
+
+let save_linear f =
+  if should_save_before_emit () then begin
+    linear_unit_info.items <- Linear_format.(Func f) :: linear_unit_info.items
+  end;
+  f
+
+let write_linear prefix =
+  if should_save_before_emit () then begin
+    let filename = Compiler_pass.(to_output_filename Scheduling ~prefix) in
+    linear_unit_info.items <- List.rev linear_unit_info.items;
+    Linear_format.save filename linear_unit_info
+  end
+
 let should_emit () =
   not (should_stop_after Compiler_pass.Scheduling)
 
@@ -46,9 +97,13 @@ let if_emit_do f x = if should_emit () then f x else ()
 let emit_begin_assembly = if_emit_do Emit.begin_assembly
 let emit_end_assembly = if_emit_do Emit.end_assembly
 let emit_data = if_emit_do Emit.data
-let emit_fundecl =
-  if_emit_do
-    (Profile.record ~accumulate:true "emit" Emit.fundecl)
+let emit_fundecl fd =
+  if should_emit() then begin
+    try
+      Profile.record ~accumulate:true "emit" Emit.fundecl fd
+    with Emitaux.Error e ->
+      raise (Error (Asm_generation(fd.Linear.fun_name, e)))
+  end
 
 let rec regalloc ~ppf_dump round fd =
   if round > 50 then
@@ -58,9 +113,9 @@ let rec regalloc ~ppf_dump round fd =
   let num_stack_slots =
     if !use_linscan then begin
       (* Linear Scan *)
-      Interval.build_intervals fd;
-      if !dump_interval then Printmach.intervals ppf_dump ();
-      Linscan.allocate_registers()
+      let intervals = Interval.build_intervals fd in
+      if !dump_interval then Printmach.intervals ppf_dump intervals;
+      Linscan.allocate_registers intervals
     end else begin
       (* Graph Coloring *)
       Interf.build_graph fd;
@@ -78,11 +133,15 @@ let rec regalloc ~ppf_dump round fd =
 
 let (++) x f = f x
 
-let compile_fundecl ~ppf_dump fd_cmm =
+let compile_fundecl ~ppf_dump ~funcnames fd_cmm =
   Proc.init ();
   Reg.reset();
   fd_cmm
-  ++ Profile.record ~accumulate:true "selection" Selection.fundecl
+  ++ Profile.record ~accumulate:true "cmm_invariants" (cmm_invariants ppf_dump)
+  ++ Profile.record ~accumulate:true "selection"
+                    (Selection.fundecl ~future_funcnames:funcnames)
+  ++ Profile.record ~accumulate:true "polling"
+                    (Polling.instrument_fundecl ~future_funcnames:funcnames)
   ++ pass_dump_if ppf_dump dump_selection "After instruction selection"
   ++ Profile.record ~accumulate:true "comballoc" Comballoc.fundecl
   ++ pass_dump_if ppf_dump dump_combine "After allocation combining"
@@ -98,19 +157,45 @@ let compile_fundecl ~ppf_dump fd_cmm =
   ++ pass_dump_if ppf_dump dump_split "After live range splitting"
   ++ Profile.record ~accumulate:true "liveness" liveness
   ++ Profile.record ~accumulate:true "regalloc" (regalloc ~ppf_dump 1)
-  ++ Profile.record ~accumulate:true "available_regs" Available_regs.fundecl
   ++ Profile.record ~accumulate:true "linearize" Linearize.fundecl
   ++ pass_dump_linear_if ppf_dump dump_linear "Linearized code"
   ++ Profile.record ~accumulate:true "scheduling" Scheduling.fundecl
   ++ pass_dump_linear_if ppf_dump dump_scheduling "After instruction scheduling"
+  ++ save_linear
   ++ emit_fundecl
 
-let compile_phrase ~ppf_dump p =
-  if !dump_cmm then fprintf ppf_dump "%a@." Printcmm.phrase p;
-  match p with
-  | Cfunction fd -> compile_fundecl ~ppf_dump fd
-  | Cdata dl -> emit_data dl
+module String = Misc.Stdlib.String
 
+let compile_data dl =
+  dl
+  ++ save_data
+  ++ emit_data
+
+let compile_phrases ~ppf_dump ps =
+  let funcnames =
+    List.fold_left (fun s p ->
+        match p with
+        | Cfunction fd -> String.Set.add fd.fun_name s
+        | Cdata _ -> s)
+      String.Set.empty ps
+  in
+  let rec compile ~funcnames ps =
+    match ps with
+    | [] -> ()
+    | p :: ps ->
+       if !dump_cmm then fprintf ppf_dump "%a@." Printcmm.phrase p;
+       match p with
+       | Cfunction fd ->
+          compile_fundecl ~ppf_dump ~funcnames fd;
+          compile ~funcnames:(String.Set.remove fd.fun_name funcnames) ps
+       | Cdata dl ->
+          compile_data dl;
+          compile ~funcnames ps
+  in
+  compile ~funcnames ps
+
+let compile_phrase ~ppf_dump p =
+  compile_phrases ~ppf_dump [p]
 
 (* For the native toplevel: generates generic functions unless
    they are already available in the process *)
@@ -122,20 +207,27 @@ let compile_genfuns ~ppf_dump f =
        | _ -> ())
     (Cmm_helpers.generic_functions true [Compilenv.current_unit_infos ()])
 
-let compile_unit asm_filename keep_asm
-      obj_filename gen =
+let compile_unit ~output_prefix ~asm_filename ~keep_asm ~obj_filename gen =
+  reset ();
   let create_asm = should_emit () &&
                    (keep_asm || not !Emitaux.binary_backend_available) in
   Emitaux.create_asm_file := create_asm;
+  let remove_asm_file () =
+    (* if [should_emit ()] is [false] then no assembly is generated,
+       so the (empty) temporary file should be deleted. *)
+    if not create_asm || not keep_asm then remove_file asm_filename
+  in
   Misc.try_finally
     ~exceptionally:(fun () -> remove_file obj_filename)
     (fun () ->
        if create_asm then Emitaux.output_channel := open_out asm_filename;
-       Misc.try_finally gen
+       Misc.try_finally
+         (fun () ->
+            gen ();
+            write_linear output_prefix)
          ~always:(fun () ->
              if create_asm then close_out !Emitaux.output_channel)
-         ~exceptionally:(fun () ->
-             if create_asm && not keep_asm then remove_file asm_filename);
+         ~exceptionally:remove_asm_file;
        if should_emit () then begin
          let assemble_result =
            Profile.record "assemble"
@@ -144,7 +236,7 @@ let compile_unit asm_filename keep_asm
          if assemble_result <> 0
          then raise(Error(Assembler_error asm_filename));
        end;
-       if create_asm && not keep_asm then remove_file asm_filename
+       remove_asm_file ()
     )
 
 let end_gen_implementation ?toplevel ~ppf_dump
@@ -152,7 +244,7 @@ let end_gen_implementation ?toplevel ~ppf_dump
   emit_begin_assembly ();
   clambda
   ++ Profile.record "cmm" Cmmgen.compunit
-  ++ Profile.record "compile_phrases" (List.iter (compile_phrase ~ppf_dump))
+  ++ Profile.record "compile_phrases" (compile_phrases ~ppf_dump)
   ++ (fun () -> ());
   (match toplevel with None -> () | Some f -> compile_genfuns ~ppf_dump f);
   (* We add explicit references to external primitive symbols.  This
@@ -170,33 +262,71 @@ let end_gen_implementation ?toplevel ~ppf_dump
 
 type middle_end =
      backend:(module Backend_intf.S)
-  -> filename:string
   -> prefixname:string
   -> ppf_dump:Format.formatter
   -> Lambda.program
   -> Clambda.with_constants
 
-let compile_implementation ?toplevel ~backend ~filename ~prefixname ~middle_end
-      ~ppf_dump (program : Lambda.program) =
-  let asmfile =
+let asm_filename output_prefix =
     if !keep_asm_file || !Emitaux.binary_backend_available
-    then prefixname ^ ext_asm
+    then output_prefix ^ ext_asm
     else Filename.temp_file "camlasm" ext_asm
-  in
-  compile_unit asmfile !keep_asm_file (prefixname ^ ext_obj)
+
+let compile_implementation ?toplevel ~backend ~prefixname ~middle_end
+      ~ppf_dump (program : Lambda.program) =
+  compile_unit ~output_prefix:prefixname
+    ~asm_filename:(asm_filename prefixname) ~keep_asm:!keep_asm_file
+    ~obj_filename:(prefixname ^ ext_obj)
     (fun () ->
       Ident.Set.iter Compilenv.require_global program.required_globals;
       let clambda_with_constants =
-        middle_end ~backend ~filename ~prefixname ~ppf_dump program
+        middle_end ~backend ~prefixname ~ppf_dump program
       in
       end_gen_implementation ?toplevel ~ppf_dump clambda_with_constants)
 
+let linear_gen_implementation filename =
+  let open Linear_format in
+  let linear_unit_info, _ = restore filename in
+  (match !Clflags.for_package, linear_unit_info.for_pack with
+   | None, None -> ()
+   | Some expected, Some saved when String.equal expected saved -> ()
+   | _, saved -> raise(Error(Mismatched_for_pack saved)));
+  let emit_item = function
+    | Data dl -> emit_data dl
+    | Func f -> emit_fundecl f
+  in
+  start_from_emit := true;
+  emit_begin_assembly ();
+  Profile.record "Emit" (List.iter emit_item) linear_unit_info.items;
+  emit_end_assembly ()
+
+let compile_implementation_linear target =
+  let output_prefix = Unit_info.prefix target in
+  compile_unit ~output_prefix
+    ~asm_filename:(asm_filename output_prefix) ~keep_asm:!keep_asm_file
+    ~obj_filename:(output_prefix ^ ext_obj)
+    (fun () ->
+      linear_gen_implementation (Unit_info.source_file target))
+
 (* Error report *)
+module Style = Misc.Style
 
 let report_error ppf = function
   | Assembler_error file ->
       fprintf ppf "Assembler error, input left in file %a"
         Location.print_filename file
+  | Mismatched_for_pack saved ->
+    let msg = function
+       | None -> Format.dprintf "without %a" Style.inline_code "-for-pack"
+       | Some s -> Format.dprintf "with %a" Style.inline_code ("-for-pack " ^ s)
+     in
+     fprintf ppf
+       "This input file cannot be compiled %t: it was generated %t."
+       (msg !Clflags.for_package) (msg saved)
+  | Asm_generation(fn, err) ->
+     fprintf ppf
+       "Error producing assembly code for function %a: %a"
+       Style.inline_code fn Emitaux.report_error err
 
 let () =
   Location.register_error_of_exn
