@@ -66,15 +66,16 @@ let env_empty = {
 
 let oper_result_type = function
     Capply ty -> ty
-  | Cextcall(_s, ty, _alloc, _) -> ty
-  | Cload (c, _) ->
-      begin match c with
+  | Cextcall(_s, ty_res, _ty_args, _alloc) -> ty_res
+  | Cload {memory_chunk} ->
+      begin match memory_chunk with
       | Word_val -> typ_val
-      | Single | Double | Double_u -> typ_float
+      | Single | Double -> typ_float
       | _ -> typ_int
       end
   | Calloc -> typ_val
   | Cstore (_c, _) -> typ_void
+  | Cdls_get -> typ_val
   | Caddi | Csubi | Cmuli | Cmulhi | Cdivi | Cmodi |
     Cand | Cor | Cxor | Clsl | Clsr | Casr |
     Ccmpi _ | Ccmpa _ | Ccmpf _ -> typ_int
@@ -85,6 +86,7 @@ let oper_result_type = function
   | Cintoffloat -> typ_int
   | Craise _ -> typ_void
   | Ccheckbound -> typ_void
+  | Copaque -> typ_val
 
 (* Infer the size in bytes of the result of an expression whose evaluation
    may be deferred (cf. [emit_parts]). *)
@@ -104,10 +106,9 @@ let size_machtype mty =
 let size_expr (env:environment) exp =
   let rec size localenv = function
       Cconst_int _ | Cconst_natint _ -> Arch.size_int
-    | Cconst_symbol _ | Cconst_pointer _ | Cconst_natpointer _ ->
+    | Cconst_symbol _ ->
         Arch.size_addr
     | Cconst_float _ -> Arch.size_float
-    | Cblockheader _ -> Arch.size_int
     | Cvar id ->
         begin try
           V.Map.find id localenv
@@ -314,10 +315,8 @@ method is_simple_expr = function
   | Cconst_natint _ -> true
   | Cconst_float _ -> true
   | Cconst_symbol _ -> true
-  | Cconst_pointer _ -> true
-  | Cconst_natpointer _ -> true
-  | Cblockheader _ -> true
   | Cvar _ -> true
+  | Creturn_addr -> true
   | Ctuple el -> List.for_all self#is_simple_expr el
   | Clet(_id, arg, body) | Clet_mut(_id, _, arg, body) ->
     self#is_simple_expr arg && self#is_simple_expr body
@@ -326,12 +325,13 @@ method is_simple_expr = function
   | Cop(op, args, _) ->
       begin match op with
         (* The following may have side effects *)
-      | Capply _ | Cextcall _ | Calloc | Cstore _ | Craise _ -> false
+      | Capply _ | Cextcall _ | Calloc | Cstore _ | Craise _ | Copaque -> false
         (* The remaining operations are simple if their args are *)
       | Cload _ | Caddi | Csubi | Cmuli | Cmulhi | Cdivi | Cmodi | Cand | Cor
       | Cxor | Clsl | Clsr | Casr | Ccmpi _ | Caddv | Cadda | Ccmpa _ | Cnegf
       | Cabsf | Caddf | Csubf | Cmulf | Cdivf | Cfloatofint | Cintoffloat
-      | Ccmpf _ | Ccheckbound -> List.for_all self#is_simple_expr args
+      | Ccmpf _ | Ccheckbound | Cdls_get ->
+          List.for_all self#is_simple_expr args
       end
   | Cassign _ | Cifthenelse _ | Cswitch _ | Ccatch _ | Cexit _
   | Ctrywith _ -> false
@@ -352,8 +352,7 @@ method effects_of exp =
   let module EC = Effect_and_coeffect in
   match exp with
   | Cconst_int _ | Cconst_natint _ | Cconst_float _ | Cconst_symbol _
-  | Cconst_pointer _ | Cconst_natpointer _ | Cblockheader _
-  | Cvar _ -> EC.none
+  | Cvar _ | Creturn_addr -> EC.none
   | Ctuple el -> EC.join_list_map el self#effects_of
   | Clet (_id, arg, body) | Clet_mut (_id, _, arg, body) ->
     EC.join (self#effects_of arg) (self#effects_of body)
@@ -366,12 +365,13 @@ method effects_of exp =
   | Cop (op, args, _) ->
     let from_op =
       match op with
-      | Capply _ | Cextcall _ -> EC.arbitrary
+      | Capply _ | Cextcall _ | Copaque -> EC.arbitrary
       | Calloc -> EC.none
       | Cstore _ -> EC.effect_only Effect.Arbitrary
       | Craise _ | Ccheckbound -> EC.effect_only Effect.Raise
-      | Cload (_, Asttypes.Immutable) -> EC.none
-      | Cload (_, Asttypes.Mutable) -> EC.coeffect_only Coeffect.Read_mutable
+      | Cload {mutability = Asttypes.Immutable} -> EC.none
+      | Cload {mutability = Asttypes.Mutable} | Cdls_get ->
+          EC.coeffect_only Coeffect.Read_mutable
       | Caddi | Csubi | Cmuli | Cmulhi | Cdivi | Cmodi | Cand | Cor | Cxor
       | Clsl | Clsr | Casr | Ccmpi _ | Caddv | Cadda | Ccmpa _ | Cnegf | Cabsf
       | Caddf | Csubf | Cmulf | Cdivf | Cfloatofint | Cintoffloat | Ccmpf _ ->
@@ -381,9 +381,18 @@ method effects_of exp =
   | Cassign _ | Cswitch _ | Ccatch _ | Cexit _ | Ctrywith _ ->
     EC.arbitrary
 
-(* Says whether an integer constant is a suitable immediate argument *)
+(* Says whether an integer constant is a suitable immediate argument for
+   the given integer operation *)
 
-method virtual is_immediate : int -> bool
+method is_immediate op n =
+  match op with
+  | Ilsl | Ilsr | Iasr -> n >= 0 && n < Arch.size_int * 8
+  | _ -> false
+
+(* Says whether an integer constant is a suitable immediate argument for
+   the given integer test *)
+
+method virtual is_immediate_test : integer_comparison -> int -> bool
 
 (* Selection of addressing modes *)
 
@@ -395,68 +404,19 @@ method virtual select_addressing :
 method select_store is_assign addr arg =
   (Istore(Word_val, addr, is_assign), arg)
 
-(* call marking methods, documented in selectgen.mli *)
-val contains_calls = ref false
-
-method mark_call =
-  contains_calls := true
-
-method mark_tailcall = ()
-
-method mark_c_tailcall = ()
-
-method mark_instr = function
-  | Iop (Icall_ind _ | Icall_imm _ | Iextcall _) ->
-      self#mark_call
-  | Iop (Itailcall_ind _ | Itailcall_imm _) ->
-      self#mark_tailcall
-  | Iop (Ialloc _) ->
-      self#mark_call (* caml_alloc*, caml_garbage_collection *)
-  | Iop (Iintop (Icheckbound _) | Iintop_imm(Icheckbound _, _)) ->
-      self#mark_c_tailcall (* caml_ml_array_bound_error *)
-  | Iraise raise_kind ->
-    begin match raise_kind with
-      | Lambda.Raise_notrace -> ()
-      | Lambda.Raise_regular
-      | Lambda.Raise_reraise ->
-          (* PR#6239 *)
-          (* caml_stash_backtrace; we #mark_call rather than
-             #mark_c_tailcall to get a good stack backtrace *)
-          self#mark_call
-    end
-  | Itrywith _ ->
-    self#mark_call
-  | _ -> ()
-
 (* Default instruction selection for operators *)
 
-method select_allocation bytes =
-  Ialloc { bytes; label_after_call_gc = None;
-           dbginfo = []; spacetime_index = 0; }
-method select_allocation_args _env = [| |]
-
-method select_checkbound () =
-  Icheckbound { spacetime_index = 0; label_after_error = None; }
-method select_checkbound_extra_args () = []
-
 method select_operation op args _dbg =
   match (op, args) with
   | (Capply _, Cconst_symbol (func, _dbg) :: rem) ->
-    let label_after = Cmm.new_label () in
-    (Icall_imm { func; label_after; }, rem)
+    (Icall_imm { func; }, rem)
   | (Capply _, _) ->
-    let label_after = Cmm.new_label () in
-    (Icall_ind { label_after; }, args)
-  | (Cextcall(func, _ty, alloc, label_after), _) ->
-    let label_after =
-      match label_after with
-      | None -> Cmm.new_label ()
-      | Some label_after -> label_after
-    in
-    Iextcall { func; alloc; label_after; }, args
-  | (Cload (chunk, _mut), [arg]) ->
-      let (addr, eloc) = self#select_addressing chunk arg in
-      (Iload(chunk, addr), [eloc])
+    (Icall_ind, args)
+  | (Cextcall(func, ty_res, ty_args, alloc), _) ->
+    Iextcall { func; alloc; ty_res; ty_args; stack_ofs = -1}, args
+  | (Cload {memory_chunk; mutability; is_atomic}, [arg]) ->
+      let (addressing_mode, eloc) = self#select_addressing memory_chunk arg in
+      (Iload {memory_chunk; addressing_mode; mutability; is_atomic}, [eloc])
   | (Cstore (chunk, init), [arg1; arg2]) ->
       let (addr, eloc) = self#select_addressing chunk arg1 in
       let is_assign =
@@ -472,7 +432,8 @@ method select_operation op args _dbg =
         (Istore(chunk, addr, is_assign), [arg2; eloc])
         (* Inversion addr/datum in Istore *)
       end
-  | (Calloc, _) -> (self#select_allocation 0), args
+  | (Cdls_get, _) -> Idls_get, args
+  | (Calloc, _) -> (Ialloc {bytes = 0; dbginfo = []}), args
   | (Caddi, _) -> self#select_arith_comm Iadd args
   | (Csubi, _) -> self#select_arith Isub args
   | (Cmuli, _) -> self#select_arith_comm Imul args
@@ -482,13 +443,14 @@ method select_operation op args _dbg =
   | (Cand, _) -> self#select_arith_comm Iand args
   | (Cor, _) -> self#select_arith_comm Ior args
   | (Cxor, _) -> self#select_arith_comm Ixor args
-  | (Clsl, _) -> self#select_shift Ilsl args
-  | (Clsr, _) -> self#select_shift Ilsr args
-  | (Casr, _) -> self#select_shift Iasr args
+  | (Clsl, _) -> self#select_arith Ilsl args
+  | (Clsr, _) -> self#select_arith Ilsr args
+  | (Casr, _) -> self#select_arith Iasr args
   | (Ccmpi comp, _) -> self#select_arith_comp (Isigned comp) args
   | (Caddv, _) -> self#select_arith_comm Iadd args
   | (Cadda, _) -> self#select_arith_comm Iadd args
   | (Ccmpa comp, _) -> self#select_arith_comp (Iunsigned comp) args
+  | (Ccmpf comp, _) -> (Icompf comp, args)
   | (Cnegf, _) -> (Inegf, args)
   | (Cabsf, _) -> (Iabsf, args)
   | (Caddf, _) -> (Iaddf, args)
@@ -498,45 +460,28 @@ method select_operation op args _dbg =
   | (Cfloatofint, _) -> (Ifloatofint, args)
   | (Cintoffloat, _) -> (Iintoffloat, args)
   | (Ccheckbound, _) ->
-    let extra_args = self#select_checkbound_extra_args () in
-    let op = self#select_checkbound () in
-    self#select_arith op (args @ extra_args)
+    self#select_arith Icheckbound args
   | _ -> Misc.fatal_error "Selection.select_oper"
 
 method private select_arith_comm op = function
-    [arg; Cconst_int (n, _)] when self#is_immediate n ->
+  | [arg; Cconst_int (n, _)] when self#is_immediate op n ->
       (Iintop_imm(op, n), [arg])
-  | [arg; Cconst_pointer (n, _)] when self#is_immediate n ->
-      (Iintop_imm(op, n), [arg])
-  | [Cconst_int (n, _); arg] when self#is_immediate n ->
-      (Iintop_imm(op, n), [arg])
-  | [Cconst_pointer (n, _); arg] when self#is_immediate n ->
+  | [Cconst_int (n, _); arg] when self#is_immediate op n ->
       (Iintop_imm(op, n), [arg])
   | args ->
       (Iintop op, args)
 
 method private select_arith op = function
-    [arg; Cconst_int (n, _)] when self#is_immediate n ->
-      (Iintop_imm(op, n), [arg])
-  | [arg; Cconst_pointer (n, _)] when self#is_immediate n ->
-      (Iintop_imm(op, n), [arg])
-  | args ->
-      (Iintop op, args)
-
-method private select_shift op = function
-    [arg; Cconst_int (n, _)] when n >= 0 && n < Arch.size_int * 8 ->
+  | [arg; Cconst_int (n, _)] when self#is_immediate op n ->
       (Iintop_imm(op, n), [arg])
   | args ->
       (Iintop op, args)
 
 method private select_arith_comp cmp = function
-    [arg; Cconst_int (n, _)] when self#is_immediate n ->
+  | [arg; Cconst_int (n, _)] when self#is_immediate (Icomp cmp) n ->
       (Iintop_imm(Icomp cmp, n), [arg])
-  | [arg; Cconst_pointer (n, _)] when self#is_immediate n ->
-      (Iintop_imm(Icomp cmp, n), [arg])
-  | [Cconst_int (n, _); arg] when self#is_immediate n ->
-      (Iintop_imm(Icomp(swap_intcomp cmp), n), [arg])
-  | [Cconst_pointer (n, _); arg] when self#is_immediate n ->
+  | [Cconst_int (n, _); arg]
+    when self#is_immediate (Icomp(swap_intcomp cmp)) n ->
       (Iintop_imm(Icomp(swap_intcomp cmp), n), [arg])
   | args ->
       (Iintop(Icomp cmp), args)
@@ -544,23 +489,19 @@ method private select_arith_comp cmp = function
 (* Instruction selection for conditionals *)
 
 method select_condition = function
-    Cop(Ccmpi cmp, [arg1; Cconst_int (n, _)], _) when self#is_immediate n ->
+  | Cop(Ccmpi cmp, [arg1; Cconst_int (n, _)], _)
+    when self#is_immediate_test (Isigned cmp) n ->
       (Iinttest_imm(Isigned cmp, n), arg1)
-  | Cop(Ccmpi cmp, [Cconst_int (n, _); arg2], _) when self#is_immediate n ->
-      (Iinttest_imm(Isigned(swap_integer_comparison cmp), n), arg2)
-  | Cop(Ccmpi cmp, [arg1; Cconst_pointer (n, _)], _) when self#is_immediate n ->
-      (Iinttest_imm(Isigned cmp, n), arg1)
-  | Cop(Ccmpi cmp, [Cconst_pointer (n, _); arg2], _) when self#is_immediate n ->
+  | Cop(Ccmpi cmp, [Cconst_int (n, _); arg2], _)
+    when self#is_immediate_test (Isigned (swap_integer_comparison cmp)) n ->
       (Iinttest_imm(Isigned(swap_integer_comparison cmp), n), arg2)
   | Cop(Ccmpi cmp, args, _) ->
       (Iinttest(Isigned cmp), Ctuple args)
-  | Cop(Ccmpa cmp, [arg1; Cconst_pointer (n, _)], _) when self#is_immediate n ->
+  | Cop(Ccmpa cmp, [arg1; Cconst_int (n, _)], _)
+    when self#is_immediate_test (Iunsigned cmp) n ->
       (Iinttest_imm(Iunsigned cmp, n), arg1)
-  | Cop(Ccmpa cmp, [arg1; Cconst_int (n, _)], _) when self#is_immediate n ->
-      (Iinttest_imm(Iunsigned cmp, n), arg1)
-  | Cop(Ccmpa cmp, [Cconst_pointer (n, _); arg2], _) when self#is_immediate n ->
-      (Iinttest_imm(Iunsigned(swap_integer_comparison cmp), n), arg2)
-  | Cop(Ccmpa cmp, [Cconst_int (n, _); arg2], _) when self#is_immediate n ->
+  | Cop(Ccmpa cmp, [Cconst_int (n, _); arg2], _)
+    when self#is_immediate_test (Iunsigned (swap_integer_comparison cmp)) n ->
       (Iinttest_imm(Iunsigned(swap_integer_comparison cmp), n), arg2)
   | Cop(Ccmpa cmp, args, _) ->
       (Iinttest(Iunsigned cmp), Ctuple args)
@@ -588,15 +529,15 @@ method insert_debug _env desc dbg arg res =
 method insert _env desc arg res =
   instr_seq <- instr_cons desc arg res instr_seq
 
-method extract_core ~end_instr =
+method extract_onto o =
   let rec extract res i =
     if i == dummy_instr
-    then res
-    else extract {i with next = res} i.next in
-  extract end_instr instr_seq
+      then res
+      else extract {i with next = res} i.next in
+    extract o instr_seq
 
 method extract =
-  self#extract_core ~end_instr:(end_instr ())
+  self#extract_onto (end_instr ())
 
 (* Insert a sequence of moves from one pseudoreg set to another. *)
 
@@ -605,7 +546,7 @@ method insert_move env src dst =
     self#insert env (Iop Imove) [|src|] [|dst|]
 
 method insert_moves env src dst =
-  for i = 0 to min (Array.length src) (Array.length dst) - 1 do
+  for i = 0 to Stdlib.Int.min (Array.length src) (Array.length dst) - 1 do
     self#insert_move env src.(i) dst.(i)
   done
 
@@ -634,20 +575,6 @@ method insert_op_debug env op dbg rs rd =
 method insert_op env op rs rd =
   self#insert_op_debug env op Debuginfo.none rs rd
 
-method emit_blockheader env n _dbg =
-  let r = self#regs_for typ_int in
-  Some(self#insert_op env (Iconst_int n) [||] r)
-
-method about_to_emit_call _env _insn _arg _dbg = None
-
-(* Prior to a function call, update the Spacetime node hole pointer hard
-   register. *)
-
-method private maybe_emit_spacetime_move env ~spacetime_reg =
-  Option.iter (fun reg ->
-      self#insert_moves env reg [| Proc.loc_spacetime_node_hole |])
-    spacetime_reg
-
 (* Add the instructions for the given expression
    at the end of the self sequence *)
 
@@ -672,14 +599,9 @@ method emit_expr (env:environment) exp =
          adding this register to the frame table would be redundant *)
       let r = self#regs_for typ_int in
       Some(self#insert_op env (Iconst_symbol n) [||] r)
-  | Cconst_pointer (n, _dbg) ->
+  | Creturn_addr ->
       let r = self#regs_for typ_int in
-      Some(self#insert_op env (Iconst_int(Nativeint.of_int n)) [||] r)
-  | Cconst_natpointer (n, _dbg) ->
-      let r = self#regs_for typ_int in
-      Some(self#insert_op env (Iconst_int n) [||] r)
-  | Cblockheader(n, dbg) ->
-      self#emit_blockheader env n dbg
+      Some(self#insert_op env Ireturn_addr [||] r)
   | Cvar v ->
       begin try
         Some(env_find v env)
@@ -726,12 +648,13 @@ method emit_expr (env:environment) exp =
           self#insert_debug env  (Iraise k) dbg rd [||];
           None
       end
-  | Cop(Ccmpf _, _, dbg) ->
-      self#emit_expr env
-        (Cifthenelse (exp,
-          dbg, Cconst_int (1, dbg),
-          dbg, Cconst_int (0, dbg),
-          dbg))
+  | Cop(Copaque, args, dbg) ->
+      begin match self#emit_parts_list env args with
+        None -> None
+      | Some (simple_args, env) ->
+         let rs = self#emit_tuple env simple_args in
+         Some (self#insert_op_debug env Iopaque dbg rs rs)
+      end
   | Cop(op, args, dbg) ->
       begin match self#emit_parts_list env args with
         None -> None
@@ -739,17 +662,13 @@ method emit_expr (env:environment) exp =
           let ty = oper_result_type op in
           let (new_op, new_args) = self#select_operation op simple_args dbg in
           match new_op with
-            Icall_ind _ ->
+            Icall_ind ->
               let r1 = self#emit_tuple env new_args in
               let rarg = Array.sub r1 1 (Array.length r1 - 1) in
               let rd = self#regs_for ty in
-              let (loc_arg, stack_ofs) = Proc.loc_arguments rarg in
-              let loc_res = Proc.loc_results rd in
-              let spacetime_reg =
-                self#about_to_emit_call env (Iop new_op) [| r1.(0) |] dbg
-              in
+              let (loc_arg, stack_ofs) = Proc.loc_arguments (Reg.typv rarg) in
+              let loc_res = Proc.loc_results (Reg.typv rd) in
               self#insert_move_args env rarg loc_arg stack_ofs;
-              self#maybe_emit_spacetime_move env ~spacetime_reg;
               self#insert_debug env (Iop new_op) dbg
                           (Array.append [|r1.(0)|] loc_arg) loc_res;
               self#insert_move_results env loc_res rd stack_ofs;
@@ -757,39 +676,31 @@ method emit_expr (env:environment) exp =
           | Icall_imm _ ->
               let r1 = self#emit_tuple env new_args in
               let rd = self#regs_for ty in
-              let (loc_arg, stack_ofs) = Proc.loc_arguments r1 in
-              let loc_res = Proc.loc_results rd in
-              let spacetime_reg =
-                self#about_to_emit_call env (Iop new_op) [| |] dbg
-              in
+              let (loc_arg, stack_ofs) = Proc.loc_arguments (Reg.typv r1) in
+              let loc_res = Proc.loc_results (Reg.typv rd) in
               self#insert_move_args env r1 loc_arg stack_ofs;
-              self#maybe_emit_spacetime_move env ~spacetime_reg;
               self#insert_debug env (Iop new_op) dbg loc_arg loc_res;
               self#insert_move_results env loc_res rd stack_ofs;
               Some rd
-          | Iextcall _ ->
-              let spacetime_reg =
-                self#about_to_emit_call env (Iop new_op) [| |] dbg
-              in
-              let (loc_arg, stack_ofs) = self#emit_extcall_args env new_args in
-              self#maybe_emit_spacetime_move env ~spacetime_reg;
+          | Iextcall r ->
+              let (loc_arg, stack_ofs) =
+                self#emit_extcall_args env r.ty_args new_args in
               let rd = self#regs_for ty in
               let loc_res =
-                self#insert_op_debug env new_op dbg
-                  loc_arg (Proc.loc_external_results rd) in
+                self#insert_op_debug env
+                  (Iextcall {r with stack_ofs = stack_ofs}) dbg
+                  loc_arg (Proc.loc_external_results (Reg.typv rd)) in
               self#insert_move_results env loc_res rd stack_ofs;
               Some rd
-          | Ialloc { bytes = _; spacetime_index; label_after_call_gc; } ->
+          | Ialloc { bytes = _; } ->
               let rd = self#regs_for typ_val in
               let bytes = size_expr env (Ctuple new_args) in
               assert (bytes mod Arch.size_addr = 0);
               let alloc_words = bytes / Arch.size_addr in
               let op =
-                Ialloc { bytes; spacetime_index; label_after_call_gc;
-                         dbginfo = [{alloc_words; alloc_dbg = dbg}] }
+                Ialloc { bytes; dbginfo = [{alloc_words; alloc_dbg = dbg}] }
               in
-              let args = self#select_allocation_args env in
-              self#insert_debug env (Iop op) dbg args rd;
+              self#insert_debug env (Iop op) dbg [||] rd;
               self#emit_stores env new_args rd;
               Some rd
           | op ->
@@ -1024,19 +935,26 @@ method private emit_tuple_not_flattened env exp_list =
 method private emit_tuple env exp_list =
   Array.concat (self#emit_tuple_not_flattened env exp_list)
 
-method emit_extcall_args env args =
+method emit_extcall_args env ty_args args =
   let args = self#emit_tuple_not_flattened env args in
-  let arg_hard_regs, stack_ofs =
-    Proc.loc_external_arguments (Array.of_list args)
-  in
-  (* Flattening [args] and [arg_hard_regs] causes parts of values split
-     across multiple registers to line up correctly, by virtue of the
-     semantics of [split_int64_for_32bit_target] in cmmgen.ml, and the
-     required semantics of [loc_external_arguments] (see proc.mli). *)
-  let args = Array.concat args in
-  let arg_hard_regs = Array.concat (Array.to_list arg_hard_regs) in
-  self#insert_move_args env args arg_hard_regs stack_ofs;
-  arg_hard_regs, stack_ofs
+  let ty_args =
+    if ty_args = [] then List.map (fun _ -> XInt) args else ty_args in
+  let locs, stack_ofs = Proc.loc_external_arguments ty_args in
+  let ty_args = Array.of_list ty_args in
+  if stack_ofs <> 0 then
+    self#insert env (Iop(Istackoffset stack_ofs)) [||] [||];
+  List.iteri
+    (fun i arg ->
+      self#insert_move_extcall_arg env ty_args.(i) arg locs.(i))
+    args;
+  Array.concat (Array.to_list locs), stack_ofs
+
+method insert_move_extcall_arg env _ty_arg src dst =
+  (* The default implementation is one or two ordinary moves.
+     (Two in the case of an int64 argument on a 32-bit platform.)
+     It can be overridden to use special move instructions,
+     for example a "32-bit move" instruction for int32 arguments. *)
+  self#insert_moves env src dst
 
 method emit_stores env data regs_addr =
   let a =
@@ -1051,7 +969,7 @@ method emit_stores env data regs_addr =
             Istore(_, _, _) ->
               for i = 0 to Array.length regs - 1 do
                 let r = regs.(i) in
-                let kind = if r.typ = Float then Double_u else Word_val in
+                let kind = if r.typ = Float then Double else Word_val in
                 self#insert env
                             (Iop(Istore(kind, !a, false)))
                             (Array.append [|r|] regs_addr) [||];
@@ -1068,7 +986,7 @@ method private emit_return (env:environment) exp =
   match self#emit_expr env exp with
     None -> ()
   | Some r ->
-      let loc = Proc.loc_results r in
+      let loc = Proc.loc_results (Reg.typv r) in
       self#insert_moves env r loc;
       self#insert env Ireturn loc [||]
 
@@ -1092,60 +1010,40 @@ method emit_tail (env:environment) exp =
       | Some(simple_args, env) ->
           let (new_op, new_args) = self#select_operation op simple_args dbg in
           match new_op with
-            Icall_ind { label_after; } ->
+            Icall_ind ->
               let r1 = self#emit_tuple env new_args in
               let rarg = Array.sub r1 1 (Array.length r1 - 1) in
-              let (loc_arg, stack_ofs) = Proc.loc_arguments rarg in
+              let (loc_arg, stack_ofs) = Proc.loc_arguments (Reg.typv rarg) in
               if stack_ofs = 0 then begin
-                let call = Iop (Itailcall_ind { label_after; }) in
-                let spacetime_reg =
-                  self#about_to_emit_call env call [| r1.(0) |] dbg
-                in
+                let call = Iop (Itailcall_ind) in
                 self#insert_moves env rarg loc_arg;
-                self#maybe_emit_spacetime_move env ~spacetime_reg;
                 self#insert_debug env call dbg
                             (Array.append [|r1.(0)|] loc_arg) [||];
               end else begin
                 let rd = self#regs_for ty in
-                let loc_res = Proc.loc_results rd in
-                let spacetime_reg =
-                  self#about_to_emit_call env (Iop new_op) [| r1.(0) |] dbg
-                in
+                let loc_res = Proc.loc_results (Reg.typv rd) in
                 self#insert_move_args env rarg loc_arg stack_ofs;
-                self#maybe_emit_spacetime_move env ~spacetime_reg;
                 self#insert_debug env (Iop new_op) dbg
                             (Array.append [|r1.(0)|] loc_arg) loc_res;
                 self#insert env (Iop(Istackoffset(-stack_ofs))) [||] [||];
                 self#insert env Ireturn loc_res [||]
               end
-          | Icall_imm { func; label_after; } ->
+          | Icall_imm { func; } ->
               let r1 = self#emit_tuple env new_args in
-              let (loc_arg, stack_ofs) = Proc.loc_arguments r1 in
+              let (loc_arg, stack_ofs) = Proc.loc_arguments (Reg.typv r1) in
               if stack_ofs = 0 then begin
-                let call = Iop (Itailcall_imm { func; label_after; }) in
-                let spacetime_reg =
-                  self#about_to_emit_call env call [| |] dbg
-                in
+                let call = Iop (Itailcall_imm { func; }) in
                 self#insert_moves env r1 loc_arg;
-                self#maybe_emit_spacetime_move env ~spacetime_reg;
                 self#insert_debug env call dbg loc_arg [||];
               end else if func = !current_function_name then begin
-                let call = Iop (Itailcall_imm { func; label_after; }) in
-                let loc_arg' = Proc.loc_parameters r1 in
-                let spacetime_reg =
-                  self#about_to_emit_call env call [| |] dbg
-                in
+                let call = Iop (Itailcall_imm { func; }) in
+                let loc_arg' = Proc.loc_parameters (Reg.typv r1) in
                 self#insert_moves env r1 loc_arg';
-                self#maybe_emit_spacetime_move env ~spacetime_reg;
                 self#insert_debug env call dbg loc_arg' [||];
               end else begin
                 let rd = self#regs_for ty in
-                let loc_res = Proc.loc_results rd in
-                let spacetime_reg =
-                  self#about_to_emit_call env (Iop new_op) [| |] dbg
-                in
+                let loc_res = Proc.loc_results (Reg.typv rd) in
                 self#insert_move_args env r1 loc_arg stack_ofs;
-                self#maybe_emit_spacetime_move env ~spacetime_reg;
                 self#insert_debug env (Iop new_op) dbg loc_arg loc_res;
                 self#insert env (Iop(Istackoffset(-stack_ofs))) [||] [||];
                 self#insert env Ireturn loc_res [||]
@@ -1215,14 +1113,14 @@ method emit_tail (env:environment) exp =
       begin match opt_r1 with
         None -> ()
       | Some r1 ->
-          let loc = Proc.loc_results r1 in
+          let loc = Proc.loc_results (Reg.typv r1) in
           self#insert_moves env r1 loc;
           self#insert env Ireturn loc [||]
       end
   | Cop _
   | Cconst_int _ | Cconst_natint _ | Cconst_float _ | Cconst_symbol _
-  | Cconst_pointer _ | Cconst_natpointer _ | Cblockheader _
   | Cvar _
+  | Creturn_addr
   | Cassign _
   | Ctuple _
   | Cexit _ ->
@@ -1233,74 +1131,44 @@ method private emit_tail_sequence env exp =
   s#emit_tail env exp;
   s#extract
 
-(* Insertion of the function prologue *)
-
-method insert_prologue _f ~loc_arg ~rarg ~spacetime_node_hole:_ ~env =
-  self#insert_moves env loc_arg rarg;
-  None
-
 (* Sequentialization of a function definition *)
 
-method initial_env () = env_empty
-
-method emit_fundecl f =
+method emit_fundecl ~future_funcnames f =
   current_function_name := f.Cmm.fun_name;
   let rargs =
     List.map
       (fun (id, ty) -> let r = self#regs_for ty in name_regs id r; r)
       f.Cmm.fun_args in
   let rarg = Array.concat rargs in
-  let loc_arg = Proc.loc_parameters rarg in
-  (* To make it easier to add the Spacetime instrumentation code, we
-     first emit the body and extract the resulting instruction sequence;
-     then we emit the prologue followed by any Spacetime instrumentation.  The
-     sequence resulting from extracting the latter (prologue + instrumentation)
-     together is then simply prepended to the body. *)
+  let loc_arg = Proc.loc_parameters (Reg.typv rarg) in
   let env =
     List.fold_right2
       (fun (id, _ty) r env -> env_add id r env)
-      f.Cmm.fun_args rargs (self#initial_env ()) in
-  let spacetime_node_hole, env =
-    if not Config.spacetime then None, env
-    else begin
-      let reg = self#regs_for typ_int in
-      let node_hole = V.create_local "spacetime_node_hole" in
-      Some (node_hole, reg), env_add (VP.create node_hole) reg env
-    end
-  in
+      f.Cmm.fun_args rargs env_empty in
   self#emit_tail env f.Cmm.fun_body;
   let body = self#extract in
   instr_seq <- dummy_instr;
-  let fun_spacetime_shape =
-    self#insert_prologue f ~loc_arg ~rarg ~spacetime_node_hole ~env
-  in
-  let body = self#extract_core ~end_instr:body in
-  instr_iter (fun instr -> self#mark_instr instr.Mach.desc) body;
+  self#insert_moves env loc_arg rarg;
+  let polled_body =
+    if Polling.requires_prologue_poll ~future_funcnames
+         ~fun_name:f.Cmm.fun_name body
+      then
+        instr_cons_debug
+          (Iop(Ipoll { return_label = None })) [||] [||] f.Cmm.fun_dbg body
+    else
+      body
+    in
+  let body_with_prologue = self#extract_onto polled_body in
   { fun_name = f.Cmm.fun_name;
     fun_args = loc_arg;
-    fun_body = body;
+    fun_body = body_with_prologue;
     fun_codegen_options = f.Cmm.fun_codegen_options;
     fun_dbg  = f.Cmm.fun_dbg;
-    fun_spacetime_shape;
+    fun_poll = f.Cmm.fun_poll;
     fun_num_stack_slots = Array.make Proc.num_register_classes 0;
-    fun_contains_calls = !contains_calls;
   }
 
 end
 
-(* Tail call criterion (estimated).  Assumes:
-- all arguments are of type "int" (always the case for OCaml function calls)
-- one extra argument representing the closure environment (conservative).
-*)
-
-let is_tail_call nargs =
-  assert (Reg.dummy.typ = Int);
-  let args = Array.make (nargs + 1) Reg.dummy in
-  let (_loc_arg, stack_ofs) = Proc.loc_arguments args in
-  stack_ofs = 0
-
-let _ =
-  Simplif.is_tail_native_heuristic := is_tail_call
-
 let reset () =
   current_function_name := ""
