@@ -125,7 +125,7 @@ end = struct
   let tmc_placeholder =
     (* we choose a placeholder whose tagged representation will be
        reconizable. *)
-    Lconst (Const_base (Const_int (0xBBBB / 2)))
+    Lambda.dummy_constant
 
   let with_placeholder constr (body : offset destination -> lambda) =
     let k_with_placeholder =
@@ -861,6 +861,12 @@ let rec choice ctx t =
     | Pignore
     | Pcompare_ints | Pcompare_floats | Pcompare_bints _
 
+    (* we don't handle effect or DLS primitives *)
+    | Prunstack | Pperform | Presume | Preperform | Pdls_get
+
+    (* we don't handle atomic primitives *)
+    | Patomic_exchange | Patomic_cas | Patomic_fetch_add | Patomic_load _
+
     (* we don't handle array indices as destinations yet *)
     | (Pmakearray _ | Pduparray _)
 
@@ -913,20 +919,43 @@ and traverse ctx = function
   | lam ->
       shallow_map (traverse ctx) lam
 
+and traverse_lfunction ctx lfun =
+  map_lfunction (traverse ctx) lfun
+
 and traverse_let outer_ctx var def =
   let inner_ctx = declare_binding outer_ctx (var, def) in
-  let bindings = traverse_binding outer_ctx inner_ctx (var, def) in
+  let bindings =
+    traverse_let_binding outer_ctx inner_ctx var def
+  in
   inner_ctx, bindings
 
 and traverse_letrec ctx bindings =
-  let ctx = List.fold_left declare_binding ctx bindings in
-  let bindings = List.concat_map (traverse_binding ctx ctx) bindings in
+  let ctx =
+    List.fold_left (fun ctx { id; def } ->
+        declare_binding ctx (id, Lfunction def)
+      ) ctx bindings
+  in
+  let bindings =
+    List.concat_map (traverse_letrec_binding ctx) bindings
+  in
   ctx, bindings
 
-and traverse_binding outer_ctx inner_ctx (var, def) =
+and traverse_let_binding outer_ctx inner_ctx var def =
   match find_candidate def with
-  | None -> [(var, traverse outer_ctx def)]
+  | None -> [ var, traverse outer_ctx def ]
   | Some lfun ->
+      let functions = make_dps_variant var inner_ctx outer_ctx lfun in
+      List.map (fun (var, lfun) -> var, Lfunction lfun) functions
+
+and traverse_letrec_binding ctx { id; def } =
+  if def.attr.tmc_candidate
+  then
+    let functions = make_dps_variant id ctx ctx def in
+    List.map (fun (id, def) -> { id; def }) functions
+  else
+    [ { id; def = traverse_lfunction ctx def } ]
+
+and make_dps_variant var inner_ctx outer_ctx (lfun : lfunction) =
   let special = Ident.Map.find var inner_ctx.specialized in
   let fun_choice = choice outer_ctx ~tail:true lfun.body in
   if fun_choice.Choice.tmc_calls = [] then
@@ -936,7 +965,7 @@ and traverse_binding outer_ctx inner_ctx (var, def) =
   let direct =
     let { kind; params; return; body = _; attr; loc } = lfun in
     let body = Choice.direct fun_choice in
-    lfunction ~kind ~params ~return ~body ~attr ~loc in
+    lfunction' ~kind ~params ~return ~body ~attr ~loc in
   let dps =
     let dst_param = {
       var = Ident.create_local "dst";
@@ -944,7 +973,7 @@ and traverse_binding outer_ctx inner_ctx (var, def) =
       loc = lfun.loc;
     } in
     let dst = { dst_param with offset = Offset (Lvar dst_param.offset) } in
-    Lambda.duplicate @@ lfunction
+    Lambda.duplicate_function @@ lfunction'
       ~kind:
         (* Support of Tupled function: see [choice_apply]. *)
         Curried
@@ -955,7 +984,7 @@ and traverse_binding outer_ctx inner_ctx (var, def) =
       ~loc:lfun.loc
   in
   let dps_var = special.dps_id in
-  [(var, direct); (dps_var, dps)]
+  [var, direct; dps_var, dps]
 
 and traverse_list ctx terms =
   List.map (traverse ctx) terms
@@ -964,6 +993,8 @@ let rewrite t =
   let ctx = { specialized = Ident.Map.empty } in
   traverse ctx t
 
+module Style = Misc.Style
+
 let () =
   Location.register_error_of_exn
     (function
@@ -971,13 +1002,16 @@ let () =
                Ambiguous_constructor_arguments
                  { explicit = false; arguments }) ->
           let print_msg ppf =
-            Format.pp_print_text ppf
-              "[@tail_mod_cons]: this constructor application may be \
-               TMC-transformed in several different ways. Please \
-               disambiguate by adding an explicit [@tailcall] \
-               attribute to the call that should be made \
-               tail-recursive, or a [@tailcall false] attribute on \
-               calls that should not be transformed."
+            Format.fprintf ppf
+              "%a:@ this@ constructor@ application@ may@ be@ \
+               TMC-transformed@ in@ several@ different@ ways.@ \
+               Please@ disambiguate@ by@ adding@ an@ explicit@ %a \
+               attribute@ to@ the@ call@ that@ should@ be@ made@ \
+               tail-recursive,@ or@ a@ %a attribute@ on@ \
+               calls@ that@ should@ not@ be@ transformed."
+              Style.inline_code "[@tail_mod_cons]"
+              Style.inline_code "[@tailcall]"
+              Style.inline_code "[@tailcall false]"
           in
           let submgs =
             let sub (info : tmc_call_information) =
@@ -993,13 +1027,14 @@ let () =
                Ambiguous_constructor_arguments
                  { explicit = true; arguments }) ->
           let print_msg ppf =
-            Format.pp_print_text ppf
-              "[@tail_mod_cons]: this constructor application may be \
-               TMC-transformed in several different ways. Only one of \
-               the arguments may become a TMC call, but several \
-               arguments contain calls that are explicitly marked as \
-               tail-recursive. Please fix the conflict by reviewing \
-               and fixing the conflicting annotations."
+            Format.fprintf ppf
+              "%a:@ this@ constructor@ application@ may@ be@ \
+               TMC-transformed@ in@ several@ different@ ways.@ Only@ one@ of@ \
+               the@ arguments@ may@ become@ a@ TMC@ call,@ but@ several@ \
+               arguments@ contain@ calls@ that@ are@ explicitly@ marked@ as@ \
+               tail-recursive.@ Please@ fix@ the@ conflict@ by@ reviewing@ \
+               and@ fixing@ the@ conflicting@ annotations."
+              Style.inline_code "[@tail_mod_cons]"
           in
           let submgs =
             let sub (info : tmc_call_information) =
