@@ -19,11 +19,11 @@
 #include <caml/fail.h>
 #include <caml/memory.h>
 #include <caml/signals.h>
-#include "unixsupport.h"
+#include "caml/unixsupport.h"
 
 #ifdef HAS_SOCKETS
 
-#include "socketaddr.h"
+#include "caml/socketaddr.h"
 #ifndef _WIN32
 #include <sys/types.h>
 #include <netdb.h>
@@ -31,69 +31,85 @@
 
 #define NETDB_BUFFER_SIZE 10000
 
-#ifdef _WIN32
-#define GETHOSTBYADDR_IS_REENTRANT 1
-#define GETHOSTBYNAME_IS_REENTRANT 1
-#endif
-
-static int entry_h_length;
-
-extern int socket_domain_table[];
-
-static value alloc_one_addr(char const *a)
+static value alloc_one_addr_4(char const *a)
 {
-  struct in_addr addr;
-#ifdef HAS_IPV6
-  struct in6_addr addr6;
-  if (entry_h_length == 16) {
-    memmove(&addr6, a, 16);
-    return alloc_inet6_addr(&addr6);
-  }
-#endif
-  memmove (&addr, a, 4);
-  return alloc_inet_addr(&addr);
+  return caml_alloc_initialized_string(4, a);
+}
+
+static value alloc_one_addr_16(char const *a)
+{
+  return caml_alloc_initialized_string(16, a);
 }
 
 static value alloc_host_entry(struct hostent *entry)
 {
+  CAMLparam0();
+  CAMLlocal4(name, aliases, addr_list, adr);
+  value (*alloc_one_addr)(char const *);
   value res;
-  value name = Val_unit, aliases = Val_unit;
-  value addr_list = Val_unit, adr = Val_unit;
 
-  Begin_roots4 (name, aliases, addr_list, adr);
-    name = caml_copy_string((char *)(entry->h_name));
-    /* PR#4043: protect against buggy implementations of gethostbyname()
-       that return a NULL pointer in h_aliases */
-    if (entry->h_aliases)
-      aliases = caml_copy_string_array((const char**)entry->h_aliases);
-    else
-      aliases = Atom(0);
-    entry_h_length = entry->h_length;
-    addr_list =
-      caml_alloc_array(alloc_one_addr, (const char**)entry->h_addr_list);
-    res = caml_alloc_small(4, 0);
-    Field(res, 0) = name;
-    Field(res, 1) = aliases;
-    switch (entry->h_addrtype) {
-    case PF_UNIX:          Field(res, 2) = Val_int(0); break;
-    case PF_INET:          Field(res, 2) = Val_int(1); break;
-    default: /*PF_INET6 */ Field(res, 2) = Val_int(2); break;
-    }
-    Field(res, 3) = addr_list;
-  End_roots();
-  return res;
+  name = caml_copy_string((char *)(entry->h_name));
+  /* PR#4043: protect against buggy implementations of gethostbyname()
+     that return a NULL pointer in h_aliases */
+  if (entry->h_aliases)
+    aliases = caml_copy_string_array((const char**)entry->h_aliases);
+  else
+    aliases = Atom(0);
+  if (entry->h_length == 16) {
+    alloc_one_addr = &alloc_one_addr_16;
+  } else {
+    CAMLassert(entry->h_length == 4);
+    alloc_one_addr = &alloc_one_addr_4;
+  }
+  addr_list =
+    caml_alloc_array(alloc_one_addr, (const char**)entry->h_addr_list);
+  res = caml_alloc_small(4, 0);
+  Field(res, 0) = name;
+  Field(res, 1) = aliases;
+  switch (entry->h_addrtype) {
+  case PF_UNIX:          Field(res, 2) = Val_int(0); break;
+  case PF_INET:          Field(res, 2) = Val_int(1); break;
+  default: /*PF_INET6 */ Field(res, 2) = Val_int(2); break;
+  }
+  Field(res, 3) = addr_list;
+  CAMLreturn(res);
 }
 
-CAMLprim value unix_gethostbyaddr(value a)
+CAMLprim value caml_unix_gethostbyaddr(value a)
 {
-  struct in_addr adr = GET_INET_ADDR(a);
+  char * adr;
+  struct in_addr in4;
   struct hostent * hp;
-#if HAS_GETHOSTBYADDR_R == 7
+  int addr_type = AF_INET;
+  socklen_t addr_len = 4;
+#if HAS_IPV6
+  struct in6_addr in6;
+  if (caml_string_length(a) == 16) {
+    addr_type = AF_INET6;
+    addr_len = 16;
+    in6 = GET_INET6_ADDR(a);
+    adr = (char *)&in6;
+  } else {
+#endif
+    in4 = GET_INET_ADDR(a);
+    adr = (char *)&in4;
+#if HAS_IPV6
+  }
+#endif
+#if !defined(HAS_GETHOSTBYADDR_R)
+#ifdef _WIN32
+  caml_enter_blocking_section();
+#endif
+  hp = gethostbyaddr(adr, addr_len, addr_type);
+#ifdef _WIN32
+  caml_leave_blocking_section();
+#endif
+#elif HAS_GETHOSTBYADDR_R == 7
   struct hostent h;
   char buffer[NETDB_BUFFER_SIZE];
   int h_errnop;
   caml_enter_blocking_section();
-  hp = gethostbyaddr_r((char *) &adr, 4, AF_INET,
+  hp = gethostbyaddr_r(adr, addr_len, addr_type,
                        &h, buffer, sizeof(buffer), &h_errnop);
   caml_leave_blocking_section();
 #elif HAS_GETHOSTBYADDR_R == 8
@@ -101,28 +117,20 @@ CAMLprim value unix_gethostbyaddr(value a)
   char buffer[NETDB_BUFFER_SIZE];
   int h_errnop, rc;
   caml_enter_blocking_section();
-  rc = gethostbyaddr_r((char *) &adr, 4, AF_INET,
+  rc = gethostbyaddr_r(adr, addr_len, addr_type,
                        &h, buffer, sizeof(buffer), &hp, &h_errnop);
   caml_leave_blocking_section();
   if (rc != 0) hp = NULL;
-#else
-#ifdef GETHOSTBYADDR_IS_REENTRANT
-  caml_enter_blocking_section();
-#endif
-  hp = gethostbyaddr((char *) &adr, 4, AF_INET);
-#ifdef GETHOSTBYADDR_IS_REENTRANT
-  caml_leave_blocking_section();
-#endif
 #endif
   if (hp == (struct hostent *) NULL) caml_raise_not_found();
   return alloc_host_entry(hp);
 }
 
-CAMLprim value unix_gethostbyname(value name)
+CAMLprim value caml_unix_gethostbyname(value name)
 {
   struct hostent * hp;
   char * hostname;
-#if HAS_GETHOSTBYNAME_R
+#ifdef HAS_GETHOSTBYNAME_R
   struct hostent h;
   char buffer[NETDB_BUFFER_SIZE];
   int err;
@@ -132,7 +140,15 @@ CAMLprim value unix_gethostbyname(value name)
 
   hostname = caml_stat_strdup(String_val(name));
 
-#if HAS_GETHOSTBYNAME_R == 5
+#if !defined(HAS_GETHOSTBYNAME_R)
+#ifdef _WIN32
+  caml_enter_blocking_section();
+#endif
+  hp = gethostbyname(hostname);
+#ifdef _WIN32
+  caml_leave_blocking_section();
+#endif
+#elif HAS_GETHOSTBYNAME_R == 5
   {
     caml_enter_blocking_section();
     hp = gethostbyname_r(hostname, &h, buffer, sizeof(buffer), &err);
@@ -146,14 +162,6 @@ CAMLprim value unix_gethostbyname(value name)
     caml_leave_blocking_section();
     if (rc != 0) hp = NULL;
   }
-#else
-#ifdef GETHOSTBYNAME_IS_REENTRANT
-  caml_enter_blocking_section();
-#endif
-  hp = gethostbyname(hostname);
-#ifdef GETHOSTBYNAME_IS_REENTRANT
-  caml_leave_blocking_section();
-#endif
 #endif
 
   caml_stat_free(hostname);
@@ -164,10 +172,10 @@ CAMLprim value unix_gethostbyname(value name)
 
 #else
 
-CAMLprim value unix_gethostbyaddr(value name)
+CAMLprim value caml_unix_gethostbyaddr(value name)
 { caml_invalid_argument("gethostbyaddr not implemented"); }
 
-CAMLprim value unix_gethostbyname(value name)
+CAMLprim value caml_unix_gethostbyname(value name)
 { caml_invalid_argument("gethostbyname not implemented"); }
 
 #endif
