@@ -27,12 +27,148 @@
 #include "caml/memory.h"
 #include "caml/mlvalues.h"
 #include "caml/signals.h"
+#include "caml/atomic_refcount.h"
 
 #define int8 caml_ba_int8
 #define uint8 caml_ba_uint8
 #define int16 caml_ba_int16
 #define uint16 caml_ba_uint16
 
+/* Half-precision floating point numbers */
+
+#if defined(__GNUC__) && defined(__aarch64__)
+
+union float16_bits { uint16_t i; _Float16 f; };
+
+Caml_inline float caml_float16_to_float(uint16 d)
+{
+  union float16_bits u;
+  u.i = d; return u.f;
+}
+
+Caml_inline uint16 caml_float_to_float16(float d)
+{
+  union float16_bits u;
+  u.f = d; return u.i;
+}
+
+#elif defined(__GNUC__) && defined(__F16C__)
+
+#include <immintrin.h>
+
+Caml_inline float caml_float16_to_float(uint16 d)
+{ return _cvtsh_ss(d); }
+
+Caml_inline uint16 caml_float_to_float16(float d)
+{ return _cvtss_sh(d, (_MM_FROUND_TO_NEAREST_INT | _MM_FROUND_NO_EXC)); }
+
+#else
+
+union float_bits {
+  uint32_t i;
+  float f;
+};
+
+/*
+ * half_to_float_fast5
+ * https://gist.github.com/rygorous/2144712
+ */
+static float caml_float16_to_float(uint16 d)
+{
+  static const union float_bits magic = { (254 - 15) << 23 };
+  static const union float_bits was_infnan = { (127 + 16) << 23 };
+
+  union float_bits o;
+
+  o.i = (d & 0x7fff) << 13;     /* exponent/mantissa bits */
+  o.f *= magic.f;               /* exponent adjust */
+  if (o.f >= was_infnan.f)      /* make sure Inf/NaN survive */
+    o.i |= 255 << 23;
+  o.i |= (d & 0x8000) << 16;    /* sign bit */
+  return o.f;
+}
+
+/*
+ * float_to_half_fast3_rtne
+ * https://gist.github.com/rygorous/2156668
+ */
+static uint16 caml_float_to_float16(float d)
+{
+  static const union float_bits f32infty = { 255 << 23 };
+  static const union float_bits f16max = { (127 + 16) << 23 };
+  static const union float_bits denorm_magic =
+    { ((127 - 15) + (23 - 10) + 1) << 23 };
+  static const uint32_t sign_mask = 0x80000000u;
+
+  union float_bits f;
+  uint16 o = 0;
+  uint32_t sign;
+
+  f.f = d;
+  sign = f.i & sign_mask;
+  f.i ^= sign;
+
+  // NOTE all the integer compares in this function can be safely
+  // compiled into signed compares since all operands are below
+  // 0x80000000. Important if you want fast straight SSE2 code
+  // (since there's no unsigned PCMPGTD).
+
+  if (f.i >= f16max.i) // result is Inf or NaN (all exponent bits set)
+    o = (f.i > f32infty.i) ? 0x7e00 : 0x7c00; // NaN->qNaN and Inf->Inf
+  else // (De)normalized number or zero
+  {
+    if (f.i < (113 << 23)) // resulting FP16 is subnormal or zero
+    {
+      // use a magic value to align our 10 mantissa bits at the bottom of
+      // the float. as long as FP addition is round-to-nearest-even this
+      // just works.
+      f.f += denorm_magic.f;
+
+      // and one integer subtract of the bias later, we have our final float!
+      o = f.i - denorm_magic.i;
+    }
+    else
+    {
+      uint32_t mant_odd = (f.i >> 13) & 1; // resulting mantissa is odd
+
+      // update exponent, rounding bias part 1
+      f.i += ((uint32_t)(15 - 127) << 23) + 0xfff;
+      // rounding bias part 2
+      f.i += mant_odd;
+      // take the bits!
+      o = f.i >> 13;
+    }
+  }
+
+  o |= sign >> 16;
+  return o;
+}
+
+#endif  /* defined(__GNUC__) && defined(__F16C__) */
+
+CAMLexport double caml_double_of_float16(intnat x)
+{
+  return (double) caml_float16_to_float((uint16) x);
+}
+
+CAMLexport intnat caml_float16_of_double(double x)
+{
+  return (intnat) caml_float_to_float16((float) x);
+}
+
+Caml_inline uint32_t caml_hash_mix_float16(uint32_t hash, uint16 d)
+{
+  /* Normalize NaNs */
+  if ((d & 0x7c00) == 0x7c00 && (d & 0x03ff) != 0) {
+    d = 0x7c01;
+  }
+  /* Normalize -0 into +0 */
+  else if (d == 0x8000) {
+    d = 0;
+  }
+  return caml_hash_mix_uint32(hash, d);
+}
+
 /* Compute the number of elements of a big array */
 
 CAMLexport uintnat caml_ba_num_elts(struct caml_ba_array * b)
@@ -53,7 +189,7 @@ CAMLexport int caml_ba_element_size[] =
   4 /*INT32*/, 8 /*INT64*/,
   sizeof(value) /*CAML_INT*/, sizeof(value) /*NATIVE_INT*/,
   8 /*COMPLEX32*/, 16 /*COMPLEX64*/,
-  1 /*CHAR*/
+  1 /*CHAR*/, 2 /*FLOAT16*/
 };
 
 /* Compute the number of bytes for the elements of a big array */
@@ -66,7 +202,7 @@ CAMLexport uintnat caml_ba_byte_size(struct caml_ba_array * b)
 
 /* Operation table for bigarrays */
 
-CAMLexport struct custom_operations caml_ba_ops = {
+CAMLexport const struct custom_operations caml_ba_ops = {
   "_bigarr02",
   caml_ba_finalize,
   caml_ba_compare,
@@ -89,31 +225,33 @@ CAMLexport value
 caml_ba_alloc(int flags, int num_dims, void * data, intnat * dim)
 {
   uintnat num_elts, asize, size;
-  int i;
+  int i, uses_resources;
   value res;
   struct caml_ba_array * b;
   intnat dimcopy[CAML_BA_MAX_NUM_DIMS];
 
   CAMLassert(num_dims >= 0 && num_dims <= CAML_BA_MAX_NUM_DIMS);
-  CAMLassert((flags & CAML_BA_KIND_MASK) <= CAML_BA_CHAR);
+  CAMLassert((flags & CAML_BA_KIND_MASK) < CAML_BA_FIRST_UNIMPLEMENTED_KIND);
   for (i = 0; i < num_dims; i++) dimcopy[i] = dim[i];
-  size = 0;
-  if (data == NULL) {
-    num_elts = 1;
-    for (i = 0; i < num_dims; i++) {
-      if (caml_umul_overflow(num_elts, dimcopy[i], &num_elts))
-        caml_raise_out_of_memory();
-    }
-    if (caml_umul_overflow(num_elts,
-                           caml_ba_element_size[flags & CAML_BA_KIND_MASK],
-                           &size))
+  num_elts = 1;
+  for (i = 0; i < num_dims; i++) {
+    if (caml_umul_overflow(num_elts, dimcopy[i], &num_elts))
       caml_raise_out_of_memory();
+  }
+  if (caml_umul_overflow(num_elts,
+                         caml_ba_element_size[flags & CAML_BA_KIND_MASK],
+                         &size))
+    caml_raise_out_of_memory();
+  if (data == NULL) {
     data = malloc(size);
     if (data == NULL && size != 0) caml_raise_out_of_memory();
     flags |= CAML_BA_MANAGED;
   }
   asize = SIZEOF_BA_ARRAY + num_dims * sizeof(intnat);
-  res = caml_alloc_custom_mem(&caml_ba_ops, asize, size);
+  uses_resources =
+    ((flags & CAML_BA_MANAGED_MASK) == CAML_BA_MANAGED)
+    && !(flags & CAML_BA_SUBARRAY);
+  res = caml_alloc_custom_mem(&caml_ba_ops, asize, uses_resources ? size : 0);
   b = Caml_ba_array_val(res);
   b->data = data;
   b->num_dims = num_dims;
@@ -154,7 +292,7 @@ CAMLexport void caml_ba_finalize(value v)
     if (b->proxy == NULL) {
       free(b->data);
     } else {
-      if (-- b->proxy->refcount == 0) {
+      if (caml_atomic_refcount_decr(&b->proxy->refcount) == 1) {
         free(b->proxy->data);
         free(b->proxy);
       }
@@ -202,10 +340,10 @@ CAMLexport int caml_ba_compare(value v1, value v2)
     } \
     return 0; \
   }
-#define DO_FLOAT_COMPARISON(type) \
-  { type * p1 = b1->data; type * p2 = b2->data; \
+#define DO_GENERIC_UNORDERED_COMPARISON(ptype, etype, conv) \
+  { ptype * p1 = b1->data; ptype * p2 = b2->data; \
     for (n = 0; n < num_elts; n++) { \
-      type e1 = *p1++; type e2 = *p2++; \
+      etype e1 = conv(*p1++); etype e2 = conv(*p2++); \
       if (e1 < e2) return -1; \
       if (e1 > e2) return 1; \
       if (e1 != e2) { \
@@ -216,8 +354,12 @@ CAMLexport int caml_ba_compare(value v1, value v2)
     } \
     return 0; \
   }
+#define DO_FLOAT_COMPARISON(type) \
+  DO_GENERIC_UNORDERED_COMPARISON(type, type, )
 
   switch (b1->flags & CAML_BA_KIND_MASK) {
+  case CAML_BA_FLOAT16:
+    DO_GENERIC_UNORDERED_COMPARISON(uint16, float, caml_float16_to_float);
   case CAML_BA_COMPLEX32:
     num_elts *= 2; /*fallthrough*/
   case CAML_BA_FLOAT32:
@@ -317,6 +459,13 @@ CAMLexport intnat caml_ba_hash(value v)
     for (n = 0; n < num_elts; n++, p++) h = caml_hash_mix_int64(h, *p);
     break;
   }
+  case CAML_BA_FLOAT16:
+  {
+    uint16 * p = b->data;
+    if (num_elts > 128) num_elts = 128;
+    for (n = 0; n < num_elts; n++, p++) h = caml_hash_mix_float16(h, *p);
+    break;
+  }
   case CAML_BA_COMPLEX32:
     num_elts *= 2;              /* fallthrough */
   case CAML_BA_FLOAT32:
@@ -394,6 +543,7 @@ CAMLexport void caml_ba_serialize(value v,
     caml_serialize_block_1(b->data, num_elts); break;
   case CAML_BA_SINT16:
   case CAML_BA_UINT16:
+  case CAML_BA_FLOAT16:
     caml_serialize_block_2(b->data, num_elts); break;
   case CAML_BA_FLOAT32:
   case CAML_BA_INT32:
@@ -462,7 +612,7 @@ CAMLexport uintnat caml_ba_deserialize(void * dst)
       caml_deserialize_error("input_value: size overflow for bigarray");
   }
   /* Determine array size in bytes.  Watch out for overflows (MPR#7765). */
-  if ((b->flags & CAML_BA_KIND_MASK) > CAML_BA_CHAR)
+  if ((b->flags & CAML_BA_KIND_MASK) >= CAML_BA_FIRST_UNIMPLEMENTED_KIND)
     caml_deserialize_error("input_value: bad bigarray kind");
   if (caml_umul_overflow(num_elts,
                          caml_ba_element_size[b->flags & CAML_BA_KIND_MASK],
@@ -480,6 +630,7 @@ CAMLexport uintnat caml_ba_deserialize(void * dst)
     caml_deserialize_block_1(b->data, num_elts); break;
   case CAML_BA_SINT16:
   case CAML_BA_UINT16:
+  case CAML_BA_FLOAT16:
     caml_deserialize_block_2(b->data, num_elts); break;
   case CAML_BA_FLOAT32:
   case CAML_BA_INT32:
@@ -553,14 +704,14 @@ static intnat caml_ba_offset(struct caml_ba_array * b, intnat * index)
 static value copy_two_doubles(double d0, double d1)
 {
   value res = caml_alloc_small(2 * Double_wosize, Double_array_tag);
-  Store_double_field(res, 0, d0);
-  Store_double_field(res, 1, d1);
+  Store_double_flat_field(res, 0, d0);
+  Store_double_flat_field(res, 1, d1);
   return res;
 }
 
 /* Generic code to read from a big array */
 
-value caml_ba_get_N(value vb, value * vind, int nind)
+value caml_ba_get_N(value vb, volatile value * vind, int nind)
 {
   struct caml_ba_array * b = Caml_ba_array_val(vb);
   intnat index[CAML_BA_MAX_NUM_DIMS];
@@ -578,6 +729,9 @@ value caml_ba_get_N(value vb, value * vind, int nind)
   switch ((b->flags) & CAML_BA_KIND_MASK) {
   default:
     CAMLassert(0);
+  case CAML_BA_FLOAT16:
+    return caml_copy_double(
+      (double) caml_float16_to_float(((uint16 *) b->data)[offset]));
   case CAML_BA_FLOAT32:
     return caml_copy_double((double) ((float *) b->data)[offset]);
   case CAML_BA_FLOAT64:
@@ -701,7 +855,8 @@ CAMLprim value caml_ba_uint8_get64(value vb, value vind)
 
 /* Generic write to a big array */
 
-static value caml_ba_set_aux(value vb, value * vind, intnat nind, value newval)
+static value caml_ba_set_aux(value vb, volatile value * vind,
+                             intnat nind, value newval)
 {
   struct caml_ba_array * b = Caml_ba_array_val(vb);
   intnat index[CAML_BA_MAX_NUM_DIMS];
@@ -719,6 +874,9 @@ static value caml_ba_set_aux(value vb, value * vind, intnat nind, value newval)
   switch (b->flags & CAML_BA_KIND_MASK) {
   default:
     CAMLassert(0);
+  case CAML_BA_FLOAT16:
+    ((uint16 *) b->data)[offset] =
+      caml_float_to_float16(Double_val(newval)); break;
   case CAML_BA_FLOAT32:
     ((float *) b->data)[offset] = Double_val(newval); break;
   case CAML_BA_FLOAT64:
@@ -740,13 +898,13 @@ static value caml_ba_set_aux(value vb, value * vind, intnat nind, value newval)
     ((intnat *) b->data)[offset] = Long_val(newval); break;
   case CAML_BA_COMPLEX32:
     { float * p = ((float *) b->data) + offset * 2;
-      p[0] = Double_field(newval, 0);
-      p[1] = Double_field(newval, 1);
+      p[0] = Double_flat_field(newval, 0);
+      p[1] = Double_flat_field(newval, 1);
       break; }
   case CAML_BA_COMPLEX64:
     { double * p = ((double *) b->data) + offset * 2;
-      p[0] = Double_field(newval, 0);
-      p[1] = Double_field(newval, 1);
+      p[0] = Double_flat_field(newval, 0);
+      p[1] = Double_flat_field(newval, 1);
       break; }
   }
   return Val_unit;
@@ -926,12 +1084,13 @@ static void caml_ba_update_proxy(struct caml_ba_array * b1,
     /* If b1 is already a proxy for a larger array, increment refcount of
        proxy */
     b2->proxy = b1->proxy;
-    ++ b1->proxy->refcount;
+    caml_atomic_refcount_incr(&b1->proxy->refcount);
   } else {
     /* Otherwise, create proxy and attach it to both b1 and b2 */
     proxy = malloc(sizeof(struct caml_ba_proxy));
     if (proxy == NULL) caml_raise_out_of_memory();
-    proxy->refcount = 2;      /* original array + sub array */
+    caml_atomic_refcount_init(&proxy->refcount, 2);
+    /* initial refcount: 2 = original array + sub array */
     proxy->data = b1->data;
     proxy->size =
       b1->flags & CAML_BA_MAPPED_FILE ? caml_ba_byte_size(b1) : 0;
@@ -976,7 +1135,8 @@ CAMLprim value caml_ba_slice(value vb, value vind)
     (char *) b->data +
     offset * caml_ba_element_size[b->flags & CAML_BA_KIND_MASK];
   /* Allocate an OCaml bigarray to hold the result */
-  res = caml_ba_alloc(b->flags, b->num_dims - num_inds, sub_data, sub_dims);
+  res = caml_ba_alloc(b->flags | CAML_BA_SUBARRAY,
+                      b->num_dims - num_inds, sub_data, sub_dims);
   /* Copy the finalization function from the original array (PR#8568) */
   Custom_ops_val(res) = Custom_ops_val(vb);
   /* Create or update proxy in case of managed bigarray */
@@ -1003,7 +1163,8 @@ CAMLprim value caml_ba_change_layout(value vb, value vlayout)
     intnat new_dim[CAML_BA_MAX_NUM_DIMS];
     unsigned int i;
     for(i = 0; i < b->num_dims; i++) new_dim[i] = b->dim[b->num_dims - i - 1];
-    res = caml_ba_alloc(flags, b->num_dims, b->data, new_dim);
+    res = caml_ba_alloc(flags | CAML_BA_SUBARRAY,
+                        b->num_dims, b->data, new_dim);
     /* Copy the finalization function from the original array (PR#8568) */
     Custom_ops_val(res) = Custom_ops_val(vb);
     caml_ba_update_proxy(b, Caml_ba_array_val(res));
@@ -1048,7 +1209,8 @@ CAMLprim value caml_ba_sub(value vb, value vofs, value vlen)
     (char *) b->data +
     ofs * mul * caml_ba_element_size[b->flags & CAML_BA_KIND_MASK];
   /* Allocate an OCaml bigarray to hold the result */
-  res = caml_ba_alloc(b->flags, b->num_dims, sub_data, b->dim);
+  res = caml_ba_alloc(b->flags | CAML_BA_SUBARRAY,
+                      b->num_dims, sub_data, b->dim);
   /* Copy the finalization function from the original array (PR#8568) */
   Custom_ops_val(res) = Custom_ops_val(vb);
   /* Doctor the changed dimension */
@@ -1128,6 +1290,12 @@ CAMLprim value caml_ba_fill(value vb, value vinit)
   switch (b->flags & CAML_BA_KIND_MASK) {
   default:
     CAMLassert(0);
+  case CAML_BA_FLOAT16: {
+    uint16 init = caml_float_to_float16(Double_val(vinit));
+    uint16 * p;
+    FILL_SCALAR_LOOP;
+    break;
+  }
   case CAML_BA_FLOAT32: {
     float init = Double_val(vinit);
     float * p;
@@ -1180,15 +1348,15 @@ CAMLprim value caml_ba_fill(value vb, value vinit)
     break;
   }
   case CAML_BA_COMPLEX32: {
-    float init0 = Double_field(vinit, 0);
-    float init1 = Double_field(vinit, 1);
+    float init0 = Double_flat_field(vinit, 0);
+    float init1 = Double_flat_field(vinit, 1);
     float * p;
     FILL_COMPLEX_LOOP;
     break;
   }
   case CAML_BA_COMPLEX64: {
-    double init0 = Double_field(vinit, 0);
-    double init1 = Double_field(vinit, 1);
+    double init0 = Double_flat_field(vinit, 0);
+    double init1 = Double_flat_field(vinit, 1);
     double * p;
     FILL_COMPLEX_LOOP;
     break;
@@ -1225,7 +1393,7 @@ CAMLprim value caml_ba_reshape(value vb, value vdim)
   if (num_elts != caml_ba_num_elts(b))
     caml_invalid_argument("Bigarray.reshape: size mismatch");
   /* Create bigarray with same data and new dimensions */
-  res = caml_ba_alloc(b->flags, num_dims, b->data, dim);
+  res = caml_ba_alloc(b->flags | CAML_BA_SUBARRAY, num_dims, b->data, dim);
   /* Copy the finalization function from the original array (PR#8568) */
   Custom_ops_val(res) = Custom_ops_val(vb);
   /* Create or update proxy in case of managed bigarray */
