@@ -17,9 +17,6 @@
 #define CAML_CUSTOM_H
 
 
-#ifndef CAML_NAME_SPACE
-#include "compatibility.h"
-#endif
 #include "mlvalues.h"
 
 struct custom_fixed_length {
@@ -48,33 +45,49 @@ struct custom_operations {
 #define custom_compare_ext_default NULL
 #define custom_fixed_length_default NULL
 
-#define Custom_ops_val(v) (*((struct custom_operations **) (v)))
+#define Custom_ops_val(v) (*((const struct custom_operations **) (v)))
 
 #ifdef __cplusplus
 extern "C" {
 #endif
 
 
-CAMLextern value caml_alloc_custom(struct custom_operations * ops,
+CAMLextern uintnat caml_custom_major_ratio;
+
+CAMLextern value caml_alloc_custom(const struct custom_operations * ops,
                                    uintnat size, /*size in bytes*/
                                    mlsize_t mem, /*resources consumed*/
                                    mlsize_t max  /*max resources*/);
 
-CAMLextern value caml_alloc_custom_mem(struct custom_operations * ops,
+/* [caml_alloc_custom_mem] allocates a custom block with dependent memory
+   (memory outside the heap that will be reclaimed when the block is
+   finalized). If [mem] is greater than [custom_minor_max_size] (see gc.mli)
+   the block is allocated directly in the major heap. */
+CAMLextern value caml_alloc_custom_mem(const struct custom_operations * ops,
                                        uintnat size, /*size in bytes*/
                                        mlsize_t mem  /*memory consumed*/);
 
-CAMLextern void caml_register_custom_operations(struct custom_operations * ops);
+CAMLextern void
+          caml_register_custom_operations(const struct custom_operations * ops);
+
+/* Return the current [max] factor for [caml_alloc_custom_mem] allocations. */
+CAMLextern mlsize_t caml_custom_get_max_major (void);
 
 /* Global variable moved to Caml_state in 4.10 */
 #define caml_compare_unordered (Caml_state_field(compare_unordered))
 
 #ifdef CAML_INTERNALS
-extern struct custom_operations * caml_find_custom_operations(char * ident);
+extern struct custom_operations *
+          caml_find_custom_operations(const char * ident);
 extern struct custom_operations *
           caml_final_custom_operations(void (*fn)(value));
 
 extern void caml_init_custom_operations(void);
+
+extern const struct custom_operations caml_nativeint_ops;
+extern const struct custom_operations caml_int32_ops;
+extern const struct custom_operations caml_int64_ops;
+extern const struct custom_operations caml_ba_ops;
 #endif /* CAML_INTERNALS */
 
 #ifdef __cplusplus
