@@ -20,7 +20,6 @@
 #include <stdio.h>
 #include <signal.h>
 #include "caml/alloc.h"
-#include "caml/domain.h"
 #include "caml/fail.h"
 #include "caml/io.h"
 #include "caml/gc.h"
@@ -31,6 +30,8 @@
 #include "caml/stack.h"
 #include "caml/roots.h"
 #include "caml/callback.h"
+#include "caml/signals.h"
+#include "caml/tsan.h"
 
 /* The globals holding predefined exceptions */
 
@@ -52,23 +53,38 @@ extern caml_generated_constant
 
 /* Exception raising */
 
-CAMLnoreturn_start
-  extern void caml_raise_exception (caml_domain_state* state, value bucket)
-CAMLnoreturn_end;
+CAMLnoret extern
+void caml_raise_exception (caml_domain_state* state, value bucket);
 
-/* Used by the stack overflow handler -> deactivate ASAN (see
-   segv_handler in signals_nat.c). */
-CAMLno_asan
 void caml_raise(value v)
 {
-  Unlock_exn();
-  if (Caml_state->exception_pointer == NULL) caml_fatal_uncaught_exception(v);
+  Caml_check_caml_state();
+  char* exception_pointer;
+  CAMLassert(!Is_exception_result(v));
+
+  caml_channel_cleanup_on_raise();
+
+  // avoid calling caml_raise recursively
+  v = caml_process_pending_actions_with_root_exn(v);
+  if (Is_exception_result(v))
+    v = Extract_exception(v);
+
+  exception_pointer = (char*)Caml_state->c_stack;
+
+  if (exception_pointer == NULL) {
+    caml_terminate_signals();
+    caml_fatal_uncaught_exception(v);
+  }
 
   while (Caml_state->local_roots != NULL &&
-         (char *) Caml_state->local_roots < Caml_state->exception_pointer) {
+         (char *) Caml_state->local_roots < exception_pointer) {
     Caml_state->local_roots = Caml_state->local_roots->next;
   }
 
+#if defined(WITH_THREAD_SANITIZER)
+  caml_tsan_exit_on_raise_c(exception_pointer);
+#endif
+
   caml_raise_exception(Caml_state, v);
 }
 
@@ -99,8 +115,7 @@ void caml_raise_with_args(value tag, int nargs, value args[])
   value bucket;
   int i;
 
-  CAMLassert(1 + nargs <= Max_young_wosize);
-  bucket = caml_alloc_small (1 + nargs, 0);
+  bucket = caml_alloc (1 + nargs, 0);
   Field(bucket, 0) = tag;
   for (i = 0; i < nargs; i++) Field(bucket, 1 + i) = args[i];
   caml_raise(bucket);
@@ -173,7 +188,7 @@ void caml_raise_sys_blocked_io(void)
   caml_raise_constant((value) caml_exn_Sys_blocked_io);
 }
 
-value caml_raise_if_exception(value res)
+CAMLexport value caml_raise_if_exception(value res)
 {
   if (Is_exception_result(res)) caml_raise(Extract_exception(res));
   return res;
@@ -182,21 +197,37 @@ value caml_raise_if_exception(value res)
 /* We use a pre-allocated exception because we can't
    do a GC before the exception is raised (lack of stack descriptors
    for the ccall to [caml_array_bound_error]).  */
-
-static const value * caml_array_bound_error_exn = NULL;
+static value array_bound_exn(void)
+{
+  static atomic_uintnat exn_cache = ATOMIC_UINTNAT_INIT(0);
+  const value* exn = (const value*)atomic_load_acquire(&exn_cache);
+  if (!exn) {
+    exn = caml_named_value("Pervasives.array_bound_error");
+    if (!exn) {
+      fprintf(stderr, "Fatal error: exception "
+        "Invalid_argument(\"index out of bounds\")\n");
+      exit(2);
+    }
+    atomic_store_release(&exn_cache, (uintnat)exn);
+  }
+  return *exn;
+}
 
 void caml_array_bound_error(void)
 {
-  if (caml_array_bound_error_exn == NULL) {
-    caml_array_bound_error_exn =
-      caml_named_value("Pervasives.array_bound_error");
-    if (caml_array_bound_error_exn == NULL) {
-      fprintf(stderr, "Fatal error: exception "
-                      "Invalid_argument(\"index out of bounds\")\n");
-      exit(2);
-    }
-  }
-  caml_raise(*caml_array_bound_error_exn);
+  caml_raise(array_bound_exn());
+}
+
+void caml_array_bound_error_asm(void)
+{
+#if defined(WITH_THREAD_SANITIZER)
+  char* exception_pointer = (char*)Caml_state->c_stack;
+  caml_tsan_exit_on_raise_c(exception_pointer);
+#endif
+
+  /* This exception is raised directly from ocamlopt-compiled OCaml,
+     not C, so we jump directly to the OCaml handler (and avoid GC) */
+  caml_raise_exception(Caml_state, array_bound_exn());
 }
 
 int caml_is_special_exception(value exn) {
