@@ -36,7 +36,7 @@
 #include "caml/mlvalues.h"
 #include "caml/misc.h"
 #include "caml/reverse.h"
-#include "caml/stacks.h"
+#include "caml/fiber.h"
 
 #if defined(HAS_LOCALE) || defined(__MINGW32__)
 
@@ -75,6 +75,10 @@
 #endif
 #endif
 
+#ifndef M_LOG2E
+#define M_LOG2E 1.44269504088896340735992468100 /* log_2 (e) */
+#endif
+
 #ifdef ARCH_ALIGN_DOUBLE
 
 CAMLexport double caml_Double_val(value val)
@@ -147,13 +151,10 @@ void caml_free_locale(void)
 
 CAMLexport value caml_copy_double(double d)
 {
+  Caml_check_caml_state();
   value res;
 
-#define Setup_for_gc
-#define Restore_after_gc
-  Alloc_small(res, Double_wosize, Double_tag);
-#undef Setup_for_gc
-#undef Restore_after_gc
+  Alloc_small(res, Double_wosize, Double_tag, Alloc_small_enter_GC);
   Store_double_val(res, d);
   return res;
 }
@@ -463,6 +464,20 @@ CAMLprim value caml_exp_float(value f)
   return caml_copy_double(exp(Double_val(f)));
 }
 
+CAMLexport double caml_exp2(double x)
+{
+#ifdef HAS_C99_FLOAT_OPS
+  return exp2(x);
+#else
+  return pow(2, x);
+#endif
+}
+
+CAMLprim value caml_exp2_float(value f)
+{
+  return caml_copy_double(caml_exp2(Double_val(f)));
+}
+
 CAMLexport double caml_trunc(double x)
 {
 #ifdef HAS_C99_FLOAT_OPS
@@ -479,7 +494,7 @@ CAMLprim value caml_trunc_float(value f)
 
 CAMLexport double caml_round(double f)
 {
-#ifdef HAS_C99_FLOAT_OPS
+#ifdef HAS_WORKING_ROUND
   return round(f);
 #else
   union { uint64_t i; double d; } u, pred_one_half; /* predecessor of 0.5 */
@@ -746,12 +761,13 @@ CAMLprim value caml_fmod_float(value f1, value f2)
 
 CAMLprim value caml_frexp_float(value f)
 {
-  CAMLparam1 (f);
-  CAMLlocal2 (res, mantissa);
+  CAMLparam0 ();
+  CAMLlocal1 (mantissa);
+  value res;
   int exponent;
 
   mantissa = caml_copy_double(frexp (Double_val(f), &exponent));
-  res = caml_alloc_tuple(2);
+  res = caml_alloc_small(2, 0);
   Field(res, 0) = mantissa;
   Field(res, 1) = Val_int(exponent);
   CAMLreturn (res);
@@ -779,16 +795,30 @@ CAMLprim value caml_log10_float(value f)
   return caml_copy_double(log10(Double_val(f)));
 }
 
+CAMLexport double caml_log2(double x)
+{
+#ifdef HAS_C99_FLOAT_OPS
+  return log2(x);
+#else
+  return log(x) * M_LOG2E;
+#endif
+}
+
+CAMLprim value caml_log2_float(value f)
+{
+  return caml_copy_double(caml_log2(Double_val(f)));
+}
+
 CAMLprim value caml_modf_float(value f)
 {
+  CAMLparam0 ();
+  CAMLlocal2 (quo, rem);
+  value res;
   double frem;
 
-  CAMLparam1 (f);
-  CAMLlocal3 (res, quo, rem);
-
   quo = caml_copy_double(modf (Double_val(f), &frem));
   rem = caml_copy_double(frem);
-  res = caml_alloc_tuple(2);
+  res = caml_alloc_small(2, 0);
   Field(res, 0) = quo;
   Field(res, 1) = rem;
   CAMLreturn (res);
@@ -799,6 +829,22 @@ CAMLprim value caml_sqrt_float(value f)
   return caml_copy_double(sqrt(Double_val(f)));
 }
 
+CAMLexport double caml_cbrt(double x)
+{
+#ifdef HAS_C99_FLOAT_OPS
+  return cbrt(x);
+#else
+  static const double third = 1.0 / 3.0;
+  double res = exp(third * log(fabs(x)));
+  return (x >= 0) ? res : -res;
+#endif
+}
+
+CAMLprim value caml_cbrt_float(value f)
+{
+  return caml_copy_double(caml_cbrt(Double_val(f)));
+}
+
 CAMLprim value caml_power_float(value f, value g)
 {
   return caml_copy_double(pow(Double_val(f), Double_val(g)));
@@ -839,16 +885,58 @@ CAMLprim value caml_asin_float(value f)
   return caml_copy_double(asin(Double_val(f)));
 }
 
+CAMLexport double caml_asinh(double x)
+{
+#ifdef HAS_C99_FLOAT_OPS
+  return asinh(x);
+#else
+  return log(x + sqrt(x * x + 1.0));
+#endif
+}
+
+CAMLprim value caml_asinh_float(value f)
+{
+  return caml_copy_double(caml_asinh(Double_val(f)));
+}
+
 CAMLprim value caml_acos_float(value f)
 {
   return caml_copy_double(acos(Double_val(f)));
 }
 
+CAMLexport double caml_acosh(double x)
+{
+#ifdef HAS_C99_FLOAT_OPS
+  return acosh(x);
+#else
+  return log(x + sqrt(x * x - 1.0));
+#endif
+}
+
+CAMLprim value caml_acosh_float(value f)
+{
+  return caml_copy_double(caml_acosh(Double_val(f)));
+}
+
 CAMLprim value caml_atan_float(value f)
 {
   return caml_copy_double(atan(Double_val(f)));
 }
 
+CAMLexport double caml_atanh(double x)
+{
+#ifdef HAS_C99_FLOAT_OPS
+  return atanh(x);
+#else
+  return 0.5 * log((1.0 + x) / (1.0 - x));
+#endif
+}
+
+CAMLprim value caml_atanh_float(value f)
+{
+  return caml_copy_double(caml_atanh(Double_val(f)));
+}
+
 CAMLprim value caml_atan2_float(value f, value g)
 {
   return caml_copy_double(atan2(Double_val(f), Double_val(g)));
@@ -921,6 +1009,53 @@ CAMLprim value caml_log1p_float(value f)
   return caml_copy_double(caml_log1p(Double_val(f)));
 }
 
+#ifndef HAS_C99_FLOAT_OPS
+Caml_inline double simple_erf(double x)
+{
+  /* This algorithm for calculating the error function is based on formula
+     7.1.26 from the "Handbook of Mathematical Functions" by Abramowitz
+     and Stegun.  The implementation using Horner's method for evaluating the
+     polynomial approximation is derived from Python code by John D. Cook. */
+  double a1 =  0.254829592, a2 = -0.284496736, a3 = 1.421413741,
+         a4 = -1.453152027, a5 =  1.061405429, p  = 0.3275911,
+         t, y;
+
+  int sign = (x >= 0) ? 1 : -1;
+  x = fabs(x);
+  t = 1.0 / (1.0 + p * x);
+  y = 1.0 - (((((a5 *t  + a4) * t) + a3) * t + a2) * t + a1) * t * exp(-x * x);
+  return sign * y;
+}
+#endif
+
+CAMLexport double caml_erf(double x)
+{
+#ifdef HAS_C99_FLOAT_OPS
+  return erf(x);
+#else
+  return simple_erf(x);
+#endif
+}
+
+CAMLprim value caml_erf_float(value f)
+{
+  return caml_copy_double(caml_erf(Double_val(f)));
+}
+
+CAMLexport double caml_erfc(double x)
+{
+#ifdef HAS_C99_FLOAT_OPS
+  return erfc(x);
+#else
+  return 1.0 - simple_erf(x);
+#endif
+}
+
+CAMLprim value caml_erfc_float(value f)
+{
+  return caml_copy_double(caml_erfc(Double_val(f)));
+}
+
 union double_as_two_int32 {
     double d;
 #if defined(ARCH_BIG_ENDIAN) || (defined(__arm__) && !defined(__ARM_EABI__))
@@ -965,16 +1100,6 @@ CAMLprim value caml_signbit_float(value f)
   return caml_signbit(Double_val(f));
 }
 
-CAMLprim value caml_neq_float(value f, value g)
-{
-  return Val_bool(Double_val(f) != Double_val(g));
-}
-
-#define DEFINE_NAN_CMP(op) (value f, value g) \
-{ \
-  return Val_bool(Double_val(f) op Double_val(g)); \
-}
-
 intnat caml_float_compare_unboxed(double f, double g)
 {
   /* If one or both of f and g is NaN, order according to the convention
@@ -989,11 +1114,15 @@ intnat caml_float_compare_unboxed(double f, double g)
   return res;
 }
 
-CAMLprim value caml_eq_float DEFINE_NAN_CMP(==)
-CAMLprim value caml_le_float DEFINE_NAN_CMP(<=)
-CAMLprim value caml_lt_float DEFINE_NAN_CMP(<)
-CAMLprim value caml_ge_float DEFINE_NAN_CMP(>=)
-CAMLprim value caml_gt_float DEFINE_NAN_CMP(>)
+#define FLOAT_CMP(op, f, g) \
+  return Val_bool(Double_val(f) op Double_val(g));
+
+CAMLprim value caml_neq_float(value f, value g) { FLOAT_CMP(!=, f, g) }
+CAMLprim value caml_eq_float(value f, value g) { FLOAT_CMP(==, f, g) }
+CAMLprim value caml_le_float(value f, value g) { FLOAT_CMP(<=, f, g) }
+CAMLprim value caml_lt_float(value f, value g) { FLOAT_CMP(<, f, g) }
+CAMLprim value caml_ge_float(value f, value g) { FLOAT_CMP(>=, f, g) }
+CAMLprim value caml_gt_float(value f, value g) { FLOAT_CMP(>, f, g) }
 
 CAMLprim value caml_float_compare(value vf, value vg)
 {
@@ -1047,23 +1176,3 @@ CAMLprim value caml_classify_float(value vd)
 {
   return caml_classify_float_unboxed(Double_val(vd));
 }
-
-/* The [caml_init_ieee_float] function should initialize floating-point hardware
-   so that it behaves as much as possible like the IEEE standard.
-   In particular, return special numbers like Infinity and NaN instead
-   of signalling exceptions.  Currently, everyone is in IEEE mode
-   at program startup, except FreeBSD prior to 4.0R. */
-
-#ifdef __FreeBSD__
-#include <osreldate.h>
-#if (__FreeBSD_version < 400017)
-#include <floatingpoint.h>
-#endif
-#endif
-
-void caml_init_ieee_floats(void)
-{
-#if defined(__FreeBSD__) && (__FreeBSD_version < 400017)
-  fpsetmask(0);
-#endif
-}
