@@ -20,6 +20,7 @@
 #include "caml/alloc.h"
 #include "caml/backtrace.h"
 #include "caml/callback.h"
+#include "caml/codefrag.h"
 #include "caml/debugger.h"
 #include "caml/fail.h"
 #include "caml/fix_code.h"
@@ -32,7 +33,10 @@
 #include "caml/mlvalues.h"
 #include "caml/prims.h"
 #include "caml/signals.h"
-#include "caml/stacks.h"
+#include "caml/fiber.h"
+#include "caml/domain.h"
+#include "caml/globroots.h"
+#include "caml/startup.h"
 #include "caml/startup_aux.h"
 
 /* Registers for the abstract machine:
@@ -40,7 +44,7 @@
         sp         the stack pointer (grows downward)
         accu       the accumulator
         env        heap-allocated environment
-        Caml_state->trapsp pointer to the current trap frame
+        Caml_state->trap_sp_off offset to the current trap frame
         extra_args number of extra arguments provided by the caller
 
 sp is a local copy of the global variable Caml_state->extern_sp. */
@@ -67,22 +71,29 @@ sp is a local copy of the global variable Caml_state->extern_sp. */
 
 /* GC interface */
 
-#undef Alloc_small_origin
-// Do call asynchronous callbacks from allocation functions
-#define Alloc_small_origin CAML_FROM_CAML
 #define Setup_for_gc \
   { sp -= 3; sp[0] = accu; sp[1] = env; sp[2] = (value)pc; \
-    Caml_state->extern_sp = sp; }
+    domain_state->current_stack->sp = sp; }
 #define Restore_after_gc \
-  { sp = Caml_state->extern_sp; accu = sp[0]; env = sp[1]; sp += 3; }
+  { sp = domain_state->current_stack->sp; accu = sp[0]; env = sp[1]; sp += 3; }
+/* Do call asynchronous callbacks from allocation functions */
+#define Enter_gc(dom_st, wosize) do {                            \
+    Setup_for_gc;                                                \
+    Alloc_small_enter_GC_flags(CAML_DO_TRACK | CAML_FROM_CAML,   \
+                               dom_st, wosize);                  \
+    Restore_after_gc;                                            \
+  } while (0)
 
 /* We store [pc+1] in the stack so that, in case of an exception, the
    first backtrace slot points to the event following the C call
    instruction. */
 #define Setup_for_c_call \
-  { sp -= 2; sp[0] = env; sp[1] = (value)(pc + 1); Caml_state->extern_sp = sp; }
+  { sp -= 2; sp[0] = env; sp[1] = (value)(pc + 1); \
+    domain_state->current_stack->sp = sp; }
 #define Restore_after_c_call \
-  { sp = Caml_state->extern_sp; env = *sp; sp += 2; }
+  { sp = domain_state->current_stack->sp; env = *sp; sp += 2; \
+    caml_update_young_limit_after_c_call(domain_state);       \
+  }
 
 /* For VM threads purposes, an event frame must look like accu + a
    C_CALL frame + a RETURN 1 frame.
@@ -93,12 +104,12 @@ sp is a local copy of the global variable Caml_state->extern_sp. */
     sp[0] = accu; /* accu */ \
     sp[1] = Val_unit; /* C_CALL frame: dummy environment */ \
     sp[2] = Val_unit; /* RETURN frame: dummy local 0 */ \
-    sp[3] = (value) pc; /* RETURN frame: saved return address */ \
+    sp[3] = (value) pc; /* RETURN frame: saved return address */  \
     sp[4] = env; /* RETURN frame: saved environment */ \
     sp[5] = Val_long(extra_args); /* RETURN frame: saved extra args */ \
-    Caml_state->extern_sp = sp; }
+    domain_state->current_stack->sp = sp; }
 #define Restore_after_event \
-  { sp = Caml_state->extern_sp; accu = sp[0]; \
+  { sp = domain_state->current_stack->sp; accu = sp[0]; \
     pc = (code_t) sp[3]; env = sp[4]; extra_args = Long_val(sp[5]); \
     sp += 6; }
 
@@ -108,9 +119,9 @@ sp is a local copy of the global variable Caml_state->extern_sp. */
    { sp -= 4; \
      sp[0] = accu; sp[1] = (value)(pc - 1); \
      sp[2] = env; sp[3] = Val_long(extra_args); \
-     Caml_state->extern_sp = sp; }
+     domain_state->current_stack->sp = sp; }
 #define Restore_after_debugger \
-   { CAMLassert(sp == Caml_state->extern_sp); \
+   { CAMLassert(sp == domain_state->current_stack->sp); \
      CAMLassert(sp[0] == accu); \
      CAMLassert(sp[2] == env); \
      sp += 4; }
@@ -124,9 +135,31 @@ sp is a local copy of the global variable Caml_state->extern_sp. */
   goto dispatch_instr
 #endif
 
-#define Check_trap_barrier \
-  if (Caml_state->trapsp >= Caml_state->trap_barrier) \
-    caml_debugger(TRAP_BARRIER, Val_unit)
+Caml_inline void check_trap_barrier_for_exception
+  (caml_domain_state* domain_state)
+{
+  if (domain_state->current_stack->id == domain_state->trap_barrier_block
+      && domain_state->trap_sp_off >= domain_state->trap_barrier_off)
+    caml_debugger(TRAP_BARRIER, Val_unit);
+}
+
+Caml_inline void check_trap_barrier_for_effect
+  (caml_domain_state* domain_state)
+{
+  if (domain_state->current_stack->id == domain_state->trap_barrier_block){
+    caml_debugger(TRAP_BARRIER, Val_unit);
+  }else{
+    struct stack_info *parent_stack
+      = domain_state->current_stack->handler->parent;
+    if (parent_stack != NULL
+        && parent_stack->id == domain_state->trap_barrier_block
+        && parent_stack->sp + 2 - Stack_high (parent_stack)
+              /* Note: +2 is the same constant as in debugger.c:552 */
+           == domain_state->trap_barrier_off){
+      caml_debugger(TRAP_BARRIER, Val_unit);
+    }
+  }
+}
 
 /* Register optimization.
    Some compilers underestimate the use of the local variables representing
@@ -206,11 +239,15 @@ sp is a local copy of the global variable Caml_state->extern_sp. */
 #endif
 
 #ifdef DEBUG
-static intnat caml_bcodcount;
+static CAMLthread_local intnat caml_bcodcount;
 #endif
 
+static value raise_unhandled_effect;
+
 /* The interpreter itself */
 
+CAMLno_tsan /* No need to TSan-instrument this (and pay a slowdown) function as
+               TSan is not supported for bytecode. */
 value caml_interprete(code_t prog, asize_t prog_size)
 {
 #ifdef PC_REG
@@ -231,12 +268,16 @@ value caml_interprete(code_t prog, asize_t prog_size)
 #endif
   value env;
   intnat extra_args;
-  struct longjmp_buffer * initial_external_raise;
-  intnat initial_sp_offset;
-  /* volatile ensures that initial_local_roots
-     will keep correct value across longjmp */
-  struct caml__roots_block * volatile initial_local_roots;
+  struct caml_exception_context * initial_external_raise;
+  int initial_stack_words;
+  intnat initial_trap_sp_off;
+  volatile value raise_exn_bucket = Val_unit;
   struct longjmp_buffer raise_buf;
+  value resume_fn, resume_arg;
+  struct stack_info* resume_tail;
+  caml_domain_state* domain_state = Caml_state;
+  struct caml_exception_context exception_ctx =
+    { &raise_buf, domain_state->local_roots, &raise_exn_bucket};
 #ifndef THREADED_CODE
   opcode_t curr_instr;
 #endif
@@ -248,38 +289,58 @@ value caml_interprete(code_t prog, asize_t prog_size)
 #endif
 
   if (prog == NULL) {           /* Interpreter is initializing */
+    static opcode_t raise_unhandled_effect_code[] = { ACC, 0, RAISE };
+    value raise_unhandled_effect_closure;
+
+    caml_register_code_fragment(
+      (char *) raise_unhandled_effect_code,
+      (char *) raise_unhandled_effect_code +
+      sizeof(raise_unhandled_effect_code),
+      DIGEST_IGNORE, NULL);
 #ifdef THREADED_CODE
     caml_init_thread_code(jumptable, Jumptbl_base);
+    caml_thread_code(raise_unhandled_effect_code,
+                     sizeof(raise_unhandled_effect_code));
 #endif
+    raise_unhandled_effect_closure = caml_alloc_small (2, Closure_tag);
+    Code_val(raise_unhandled_effect_closure) =
+      (code_t)raise_unhandled_effect_code;
+    Closinfo_val(raise_unhandled_effect_closure) = Make_closinfo(0, 2);
+    raise_unhandled_effect = raise_unhandled_effect_closure;
+    caml_register_generational_global_root(&raise_unhandled_effect);
+    caml_global_data = Val_unit;
+    caml_register_generational_global_root(&caml_global_data);
+    caml_init_callbacks();
     return Val_unit;
   }
 
 #if defined(THREADED_CODE) && defined(ARCH_SIXTYFOUR) && !defined(ARCH_CODE32)
   jumptbl_base = Jumptbl_base;
 #endif
-  initial_local_roots = Caml_state->local_roots;
-  initial_sp_offset =
-    (char *) Caml_state->stack_high - (char *) Caml_state->extern_sp;
-  initial_external_raise = Caml_state->external_raise;
-  caml_callback_depth++;
+  initial_trap_sp_off = domain_state->trap_sp_off;
+  initial_stack_words =
+    Stack_high(domain_state->current_stack) - domain_state->current_stack->sp;
+  initial_external_raise = domain_state->external_raise;
 
   if (sigsetjmp(raise_buf.buf, 0)) {
-    Caml_state->local_roots = initial_local_roots;
-    sp = Caml_state->extern_sp;
-    accu = Caml_state->exn_bucket;
+    /* no non-volatile local variables read here */
+    sp = domain_state->current_stack->sp;
+    accu = raise_exn_bucket;
 
-    Check_trap_barrier;
-    if (Caml_state->backtrace_active) {
-      /* pc has already been pushed on the stack when calling the C
+    check_trap_barrier_for_exception (domain_state);
+    if (domain_state->backtrace_active) {
+         /* pc has already been pushed on the stack when calling the C
          function that raised the exception. No need to push it again
          here. */
       caml_stash_backtrace(accu, sp, 0);
     }
     goto raise_notrace;
   }
-  Caml_state->external_raise = &raise_buf;
+  domain_state->external_raise = &exception_ctx;
 
-  sp = Caml_state->extern_sp;
+  domain_state->trap_sp_off = 1;
+
+  sp = domain_state->current_stack->sp;
   pc = prog;
   extra_args = 0;
   env = Atom(0);
@@ -289,8 +350,8 @@ value caml_interprete(code_t prog, asize_t prog_size)
 #ifdef DEBUG
  next_instr:
   if (caml_icount-- == 0) caml_stop_here ();
-  CAMLassert(sp >= Caml_state->stack_low);
-  CAMLassert(sp <= Caml_state->stack_high);
+  CAMLassert(Stack_base(domain_state->current_stack) <= sp);
+  CAMLassert(sp <= Stack_high(domain_state->current_stack));
 #endif
   goto *(void *)(jumptbl_base + *pc++); /* Jump to the first instruction */
 #else
@@ -298,18 +359,20 @@ value caml_interprete(code_t prog, asize_t prog_size)
 #ifdef DEBUG
     caml_bcodcount++;
     if (caml_icount-- == 0) caml_stop_here ();
-    if (caml_trace_level>1) printf("\n##%" ARCH_INTNAT_PRINTF_FORMAT "d\n",
-                                   caml_bcodcount);
-    if (caml_trace_level>0) caml_disasm_instr(pc);
-    if (caml_trace_level>1) {
+    if (caml_params->trace_level>1)
+      printf("\n##%" ARCH_INTNAT_PRINTF_FORMAT "d\n", caml_bcodcount);
+    if (caml_params->trace_level>0) caml_disasm_instr(pc);
+    if (caml_params->event_trace>0) caml_event_trace(pc);
+    if (caml_params->trace_level>1) {
       printf("env=");
       caml_trace_value_file(env,prog,prog_size,stdout);
       putchar('\n');
       caml_trace_accu_sp_file(accu,sp,prog,prog_size,stdout);
       fflush(stdout);
     };
-    CAMLassert(sp >= Caml_state->stack_low);
-    CAMLassert(sp <= Caml_state->stack_high);
+    CAMLassert(Stack_base(domain_state->current_stack) <= sp);
+    CAMLassert(sp <= Stack_high(domain_state->current_stack));
+
 #endif
     curr_instr = *pc++;
 
@@ -508,14 +571,41 @@ value caml_interprete(code_t prog, asize_t prog_size)
         extra_args--;
         pc = Code_val(accu);
         env = accu;
+        Next;
       } else {
+        goto do_return;
+      }
+    }
+
+    do_return:
+      if (sp == Stack_high(domain_state->current_stack)) {
+        /* return to parent stack */
+        struct stack_info* old_stack = domain_state->current_stack;
+        struct stack_info* parent_stack = Stack_parent(old_stack);
+        value hval = Stack_handle_value(old_stack);
+        CAMLassert(parent_stack != NULL);
+
+        domain_state->current_stack = parent_stack;
+        sp = domain_state->current_stack->sp;
+        caml_free_stack(old_stack);
+
+        domain_state->trap_sp_off = Long_val(sp[0]);
+        extra_args = Long_val(sp[1]);
+        sp++;
+        sp[0] = accu;
+
+        accu = hval;
+        pc = Code_val(accu);
+        env = accu;
+        goto check_stacks;
+      } else {
+        /* return to callee, no stack switching */
         pc = (code_t)(sp[0]);
         env = sp[1];
         extra_args = Long_val(sp[2]);
         sp += 3;
       }
       Next;
-    }
 
     Instruct(RESTART): {
       int num_args = Wosize_val(env) - 3;
@@ -531,21 +621,18 @@ value caml_interprete(code_t prog, asize_t prog_size)
       int required = *pc++;
       if (extra_args >= required) {
         extra_args -= required;
+        Next;
       } else {
         mlsize_t num_args, i;
         num_args = 1 + extra_args; /* arg1 + extra args */
-        Alloc_small(accu, num_args + 3, Closure_tag);
+        Alloc_small(accu, num_args + 3, Closure_tag, Enter_gc);
         Field(accu, 2) = env;
         for (i = 0; i < num_args; i++) Field(accu, i + 3) = sp[i];
         Code_val(accu) = pc - 3; /* Point to the preceding RESTART instr. */
         Closinfo_val(accu) = Make_closinfo(0, 2);
         sp += num_args;
-        pc = (code_t)(sp[0]);
-        env = sp[1];
-        extra_args = Long_val(sp[2]);
-        sp += 3;
+        goto do_return;
       }
-      Next;
     }
 
     Instruct(CLOSURE): {
@@ -554,7 +641,7 @@ value caml_interprete(code_t prog, asize_t prog_size)
       if (nvars > 0) *--sp = accu;
       if (nvars <= Max_young_wosize - 2) {
         /* nvars + 2 <= Max_young_wosize, can allocate in minor heap */
-        Alloc_small(accu, 2 + nvars, Closure_tag);
+        Alloc_small(accu, 2 + nvars, Closure_tag, Enter_gc);
         for (i = 0; i < nvars; i++) Field(accu, i + 2) = sp[i];
       } else {
         /* PR#6385: must allocate in major heap */
@@ -578,10 +665,10 @@ value caml_interprete(code_t prog, asize_t prog_size)
       mlsize_t envofs = nfuncs * 3 - 1;
       mlsize_t blksize = envofs + nvars;
       int i;
-      value * p;
+      volatile value * p;
       if (nvars > 0) *--sp = accu;
       if (blksize <= Max_young_wosize) {
-        Alloc_small(accu, blksize, Closure_tag);
+        Alloc_small(accu, blksize, Closure_tag, Enter_gc);
         p = &Field(accu, envofs);
         for (i = 0; i < nvars; i++, p++) *p = sp[i];
       } else {
@@ -600,7 +687,7 @@ value caml_interprete(code_t prog, asize_t prog_size)
       *p++ = (value) (pc + pc[0]);
       *p++ = Make_closinfo(0, envofs);
       for (i = 1; i < nfuncs; i++) {
-        *p++ = Make_header(i * 3, Infix_tag, Caml_white); /* color irrelevant */
+        *p++ = Make_header(i * 3, Infix_tag, 0); /* color irrelevant */
         *--sp = (value) p;
         *p++ = (value) (pc + pc[i]);
         envofs -= 3;
@@ -650,11 +737,12 @@ value caml_interprete(code_t prog, asize_t prog_size)
       Next;
     }
 
-    Instruct(SETGLOBAL):
+    Instruct(SETGLOBAL):  {
       caml_modify(&Field(caml_global_data, *pc), accu);
       accu = Val_unit;
       pc++;
       Next;
+    }
 
 /* Allocation of blocks */
 
@@ -676,7 +764,7 @@ value caml_interprete(code_t prog, asize_t prog_size)
       mlsize_t i;
       value block;
       if (wosize <= Max_young_wosize) {
-        Alloc_small(block, wosize, tag);
+        Alloc_small(block, wosize, tag, Enter_gc);
         Field(block, 0) = accu;
         for (i = 1; i < wosize; i++) Field(block, i) = *sp++;
       } else {
@@ -690,7 +778,7 @@ value caml_interprete(code_t prog, asize_t prog_size)
     Instruct(MAKEBLOCK1): {
       tag_t tag = *pc++;
       value block;
-      Alloc_small(block, 1, tag);
+      Alloc_small(block, 1, tag, Enter_gc);
       Field(block, 0) = accu;
       accu = block;
       Next;
@@ -698,7 +786,7 @@ value caml_interprete(code_t prog, asize_t prog_size)
     Instruct(MAKEBLOCK2): {
       tag_t tag = *pc++;
       value block;
-      Alloc_small(block, 2, tag);
+      Alloc_small(block, 2, tag, Enter_gc);
       Field(block, 0) = accu;
       Field(block, 1) = sp[0];
       sp += 1;
@@ -708,7 +796,7 @@ value caml_interprete(code_t prog, asize_t prog_size)
     Instruct(MAKEBLOCK3): {
       tag_t tag = *pc++;
       value block;
-      Alloc_small(block, 3, tag);
+      Alloc_small(block, 3, tag, Enter_gc);
       Field(block, 0) = accu;
       Field(block, 1) = sp[0];
       Field(block, 2) = sp[1];
@@ -721,7 +809,7 @@ value caml_interprete(code_t prog, asize_t prog_size)
       mlsize_t i;
       value block;
       if (size <= Max_young_wosize / Double_wosize) {
-        Alloc_small(block, size * Double_wosize, Double_array_tag);
+        Alloc_small(block, size * Double_wosize, Double_array_tag, Enter_gc);
       } else {
         block = caml_alloc_shr(size * Double_wosize, Double_array_tag);
       }
@@ -748,7 +836,7 @@ value caml_interprete(code_t prog, asize_t prog_size)
       accu = Field(accu, *pc); pc++; Next;
     Instruct(GETFLOATFIELD): {
       double d = Double_flat_field(accu, *pc++);
-      Alloc_small(accu, Double_wosize, Double_tag);
+      Alloc_small(accu, Double_wosize, Double_tag, Enter_gc);
       Store_double_val(accu, d);
       Next;
     }
@@ -847,77 +935,102 @@ value caml_interprete(code_t prog, asize_t prog_size)
     Instruct(PUSHTRAP):
       sp -= 4;
       Trap_pc(sp) = pc + *pc;
-      Trap_link_offset(sp) = Val_long(Caml_state->trapsp - sp);
+      Trap_link(sp) = Val_long(domain_state->trap_sp_off);
       sp[2] = env;
       sp[3] = Val_long(extra_args);
-      Caml_state->trapsp = sp;
+      domain_state->trap_sp_off = sp - Stack_high(domain_state->current_stack);
       pc++;
       Next;
 
     Instruct(POPTRAP):
-      if (caml_something_to_do) {
+      if (Caml_check_gc_interrupt(domain_state)) {
         /* We must check here so that if a signal is pending and its
            handler triggers an exception, the exception is trapped
            by the current try...with, not the enclosing one. */
         pc--; /* restart the POPTRAP after processing the signal */
-        goto process_actions;
+        goto process_signal;
       }
-      Caml_state->trapsp = sp + Long_val(Trap_link_offset(sp));
+      domain_state->trap_sp_off = Long_val(Trap_link(sp));
       sp += 4;
       Next;
 
     Instruct(RAISE_NOTRACE):
-      Check_trap_barrier;
+      check_trap_barrier_for_exception (domain_state);
       goto raise_notrace;
 
     Instruct(RERAISE):
-      Check_trap_barrier;
-      if (Caml_state->backtrace_active) {
+      check_trap_barrier_for_exception (domain_state);
+      if (domain_state->backtrace_active) {
         *--sp = (value)(pc - 1);
         caml_stash_backtrace(accu, sp, 1);
       }
       goto raise_notrace;
 
     Instruct(RAISE):
-      Check_trap_barrier;
-      if (Caml_state->backtrace_active) {
+    raise_exception:
+      check_trap_barrier_for_exception (domain_state);
+      if (domain_state->backtrace_active) {
         *--sp = (value)(pc - 1);
         caml_stash_backtrace(accu, sp, 0);
       }
     raise_notrace:
-      if ((char *) Caml_state->trapsp
-          >= (char *) Caml_state->stack_high - initial_sp_offset) {
-        Caml_state->external_raise = initial_external_raise;
-        Caml_state->extern_sp = (value *) ((char *) Caml_state->stack_high
-                                    - initial_sp_offset);
-        caml_callback_depth--;
-        return Make_exception_result(accu);
+      if (domain_state->trap_sp_off > 0) {
+        if (Stack_parent(domain_state->current_stack) == NULL) {
+          domain_state->external_raise = initial_external_raise;
+          domain_state->trap_sp_off = initial_trap_sp_off;
+          domain_state->current_stack->sp =
+            Stack_high(domain_state->current_stack) - initial_stack_words ;
+          return Make_exception_result(accu);
+        } else {
+          struct stack_info* old_stack = domain_state->current_stack;
+          struct stack_info* parent_stack = Stack_parent(old_stack);
+          value hexn = Stack_handle_exception(old_stack);
+          old_stack->sp = sp;
+          domain_state->current_stack = parent_stack;
+          sp = domain_state->current_stack->sp;
+          caml_free_stack(old_stack);
+
+          domain_state->trap_sp_off = Long_val(sp[0]);
+          extra_args = Long_val(sp[1]);
+          sp++;
+          sp[0] = accu;
+
+          accu = hexn;
+          pc = Code_val(accu);
+          env = accu;
+          goto check_stacks;
+        }
+      } else {
+        sp =
+           Stack_high(domain_state->current_stack) + domain_state->trap_sp_off;
+        pc = Trap_pc(sp);
+        domain_state->trap_sp_off = Long_val(Trap_link(sp));
+        env = sp[2];
+        extra_args = Long_val(sp[3]);
+        sp += 4;
       }
-      sp = Caml_state->trapsp;
-      pc = Trap_pc(sp);
-      Caml_state->trapsp = sp + Long_val(Trap_link_offset(sp));
-      env = sp[2];
-      extra_args = Long_val(sp[3]);
-      sp += 4;
       Next;
 
 /* Stack checks */
 
     check_stacks:
-      if (sp < Caml_state->stack_threshold) {
-        Caml_state->extern_sp = sp;
-        caml_realloc_stack(Stack_threshold / sizeof(value));
-        sp = Caml_state->extern_sp;
+      if (sp < Stack_threshold_ptr(domain_state->current_stack)) {
+        domain_state->current_stack->sp = sp;
+        if (!caml_try_realloc_stack(Stack_threshold / sizeof(value))) {
+          Setup_for_c_call; caml_raise_stack_overflow();
+        }
+        sp = domain_state->current_stack->sp;
       }
       /* Fall through CHECK_SIGNALS */
 
 /* Signal handling */
 
     Instruct(CHECK_SIGNALS):    /* accu not preserved */
-      if (caml_something_to_do) goto process_actions;
+      if (Caml_check_gc_interrupt(domain_state))
+        goto process_signal;
       Next;
 
-    process_actions:
+    process_signal:
       Setup_for_event;
       caml_process_pending_actions();
       Restore_after_event;
@@ -927,34 +1040,34 @@ value caml_interprete(code_t prog, asize_t prog_size)
 
     Instruct(C_CALL1):
       Setup_for_c_call;
-      accu = Primitive(*pc)(accu);
+      accu = Primitive1(*pc)(accu);
       Restore_after_c_call;
       pc++;
       Next;
     Instruct(C_CALL2):
       Setup_for_c_call;
-      accu = Primitive(*pc)(accu, sp[2]);
+      accu = Primitive2(*pc)(accu, sp[2]);
       Restore_after_c_call;
       sp += 1;
       pc++;
       Next;
     Instruct(C_CALL3):
       Setup_for_c_call;
-      accu = Primitive(*pc)(accu, sp[2], sp[3]);
+      accu = Primitive3(*pc)(accu, sp[2], sp[3]);
       Restore_after_c_call;
       sp += 2;
       pc++;
       Next;
     Instruct(C_CALL4):
       Setup_for_c_call;
-      accu = Primitive(*pc)(accu, sp[2], sp[3], sp[4]);
+      accu = Primitive4(*pc)(accu, sp[2], sp[3], sp[4]);
       Restore_after_c_call;
       sp += 3;
       pc++;
       Next;
     Instruct(C_CALL5):
       Setup_for_c_call;
-      accu = Primitive(*pc)(accu, sp[2], sp[3], sp[4], sp[5]);
+      accu = Primitive5(*pc)(accu, sp[2], sp[3], sp[4], sp[5]);
       Restore_after_c_call;
       sp += 4;
       pc++;
@@ -963,7 +1076,7 @@ value caml_interprete(code_t prog, asize_t prog_size)
       int nargs = *pc++;
       *--sp = accu;
       Setup_for_c_call;
-      accu = Primitive(*pc)(sp + 2, nargs);
+      accu = PrimitiveN(*pc)(sp + 2, nargs);
       Restore_after_c_call;
       sp += nargs;
       pc++;
@@ -1101,7 +1214,9 @@ value caml_interprete(code_t prog, asize_t prog_size)
 #endif
       *--sp = accu;
       accu = Val_int(*pc++);
-      ofs = *pc & Field(meths,1);
+      /* We use relaxed atomic accesses to avoid racing with other domains
+         updating the cache */
+      ofs = atomic_load_relaxed((_Atomic opcode_t *)pc) & Field(meths,1);
       if (*(value*)(((char*)&Field(meths,3)) + ofs) == accu) {
 #ifdef CAML_TEST_CACHE
         hits++;
@@ -1116,7 +1231,7 @@ value caml_interprete(code_t prog, asize_t prog_size)
           if (accu < Field(meths,mi)) hi = mi-2;
           else li = mi;
         }
-        *pc = (li-3)*sizeof(value);
+        atomic_store_relaxed((_Atomic opcode_t *)pc, (li-3)*sizeof(value));
         accu = Field (meths, li-1);
       }
       pc++;
@@ -1145,9 +1260,9 @@ value caml_interprete(code_t prog, asize_t prog_size)
 /* Debugging and machine control */
 
     Instruct(STOP):
-      Caml_state->external_raise = initial_external_raise;
-      Caml_state->extern_sp = sp;
-      caml_callback_depth--;
+      domain_state->external_raise = initial_external_raise;
+      domain_state->trap_sp_off = initial_trap_sp_off;
+      domain_state->current_stack->sp = sp;
       return accu;
 
     Instruct(EVENT):
@@ -1164,9 +1279,142 @@ value caml_interprete(code_t prog, asize_t prog_size)
       Restore_after_debugger;
       Restart_curr_instr;
 
+/* Context switching */
+
+    Instruct(RESUME):
+      resume_fn = sp[0];
+      resume_arg = sp[1];
+      resume_tail = Ptr_val(sp[2]);
+      sp -= 2;
+      sp[0] = Val_long(domain_state->trap_sp_off);
+      sp[1] = Val_long(0);
+      sp[2] = (value)pc;
+      sp[3] = env;
+      sp[4] = Val_long(extra_args);
+      goto do_resume;
+
+do_resume: {
+      struct stack_info* stk = Ptr_val(accu);
+      if (stk == NULL) {
+        Setup_for_c_call;
+        caml_raise_continuation_already_resumed();
+      }
+      if (resume_tail == NULL) {
+        resume_tail = stk;
+      }
+      Stack_parent(resume_tail) = Caml_state->current_stack;
+
+      domain_state->current_stack->sp = sp;
+      domain_state->current_stack = stk;
+      sp = domain_state->current_stack->sp;
+
+      domain_state->trap_sp_off = Long_val(sp[0]);
+      sp[0] = resume_arg;
+      accu = resume_fn;
+      pc = Code_val(accu);
+      env = accu;
+      extra_args = 0;
+      goto check_stacks;
+    }
+
+    Instruct(RESUMETERM):
+      resume_fn = sp[0];
+      resume_arg = sp[1];
+      resume_tail = Ptr_val(sp[2]);
+      sp = sp + *pc - 2;
+      sp[0] = Val_long(domain_state->trap_sp_off);
+      sp[1] = Val_long(extra_args);
+      goto do_resume;
+
+
+    Instruct(PERFORM): {
+      value cont;
+      struct stack_info* old_stack = domain_state->current_stack;
+      struct stack_info* parent_stack = Stack_parent(old_stack);
+
+      check_trap_barrier_for_effect (domain_state);
+      if (parent_stack == NULL) {
+        Setup_for_c_call;
+        accu = caml_make_unhandled_effect_exn(accu);
+        Restore_after_c_call;
+        goto raise_exception;
+      }
+
+      Alloc_small(cont, 2, Cont_tag, Enter_gc);
+
+      sp -= 4;
+      sp[0] = Val_long(domain_state->trap_sp_off);
+      sp[1] = (value)pc;
+      sp[2] = env;
+      sp[3] = Val_long(extra_args);
+
+      old_stack->sp = sp;
+      domain_state->current_stack = parent_stack;
+      sp = parent_stack->sp;
+      Stack_parent(old_stack) = NULL;
+      Field(cont, 0) = Val_ptr(old_stack);
+      Field(cont, 1) = Val_long(0);
+
+      domain_state->trap_sp_off = Long_val(sp[0]);
+      extra_args = Long_val(sp[1]);
+      sp--;
+      sp[0] = accu;
+      sp[1] = cont;
+      sp[2] = Val_ptr(old_stack);
+      accu = Stack_handle_effect(old_stack);
+      pc = Code_val(accu);
+      env = accu;
+      extra_args += 2;
+      goto check_stacks;
+    }
+
+    Instruct(REPERFORMTERM): {
+      value eff = accu;
+      value cont = sp[0];
+      struct stack_info* cont_tail = Ptr_val(sp[1]);
+      struct stack_info* self = domain_state->current_stack;
+      struct stack_info* parent = Stack_parent(domain_state->current_stack);
+
+      check_trap_barrier_for_effect (domain_state);
+      sp = sp + *pc - 2;
+      sp[0] = Val_long(domain_state->trap_sp_off);
+      sp[1] = Val_long(extra_args);
+
+      if (parent == NULL) {
+        Setup_for_c_call;
+        resume_arg = caml_make_unhandled_effect_exn(eff);
+        accu = caml_continuation_use(cont);
+        Restore_after_c_call;
+        resume_fn = raise_unhandled_effect;
+        resume_tail = cont_tail;
+
+        goto do_resume;
+      }
+
+      self->sp = sp;
+      domain_state->current_stack = parent;
+      sp = parent->sp;
+
+      CAMLassert(Stack_parent(cont_tail) == NULL);
+      Stack_parent(self) = NULL;
+      Stack_parent(cont_tail) = self;
+
+      domain_state->trap_sp_off = Long_val(sp[0]);
+      extra_args = Long_val(sp[1]);
+      sp--;
+      sp[0] = eff;
+      sp[1] = cont;
+      sp[2] = Val_ptr(self);
+      accu = Stack_handle_effect(self);
+      pc = Code_val(accu);
+      env = accu;
+      extra_args += 2;
+      goto check_stacks;
+    }
+
 #ifndef THREADED_CODE
     default:
-#if _MSC_VER >= 1200
+#ifdef _MSC_VER
       __assume(0);
 #else
       caml_fatal_error("bad opcode (%"
