@@ -20,17 +20,15 @@
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
-#include "../config/s.h"
+#include "caml/s.h"
 #ifdef HAS_UNISTD
 #include <unistd.h>
 #endif
 #include <fcntl.h>
 #include <sys/types.h>
 #include <sys/stat.h>
-#include "../byterun/caml/mlvalues.h"
-#include "../byterun/caml/exec.h"
-
-char * default_runtime_path = RUNTIME_NAME;
+#include "caml/mlvalues.h"
+#include "caml/exec.h"
 
 #ifndef MAXPATHLEN
 #define MAXPATHLEN 1024
@@ -155,17 +153,16 @@ static char * read_runtime_path(int fd)
     } else if (path_size > 0)
       ofs += read_size(buffer + 4);
   }
-  if (path_size == 0) return default_runtime_path;
+  if (path_size == 0) return NULL;
   if (path_size >= MAXPATHLEN) return NULL;
   lseek(fd, -ofs, SEEK_END);
   if (read(fd, runtime_path, path_size) != path_size) return NULL;
-  runtime_path[path_size - 1] = 0;
   return runtime_path;
 }
 
 static void errwrite(char * msg)
 {
-  write(2, msg, strlen(msg));
+  fputs(msg, stderr);
 }
 
 #ifndef O_BINARY
