@@ -29,14 +29,14 @@
    the [Marshal.from_*] functions is given as ['a], but this is
    misleading: the returned OCaml value does not possess type ['a]
    for all ['a]; it has one, unique type which cannot be determined
-   at compile-type.  The programmer should explicitly give the expected
+   at compile-time.  The programmer should explicitly give the expected
    type of the returned value, using the following syntax:
    - [(Marshal.from_channel chan : type)].
    Anything can happen at run-time if the object in the file does not
    belong to the given type.
 
    Values of extensible variant types, for example exceptions (of
-   extensible type [exn]), returned by the unmarhsaller should not be
+   extensible type [exn]), returned by the unmarshaller should not be
    pattern-matched over through [match ... with] or [try ... with],
    because unmarshalling does not preserve the information required for
    matching their constructors. Structural equalities with other
@@ -56,6 +56,7 @@ type extern_flags =
     No_sharing                          (** Don't preserve sharing *)
   | Closures                            (** Send function closures *)
   | Compat_32                           (** Ensure 32-bit compatibility *)
+
 (** The flags to the [Marshal.to_*] functions below. *)
 
 val to_channel : out_channel -> 'a -> extern_flags list -> unit
@@ -97,27 +98,28 @@ val to_channel : out_channel -> 'a -> extern_flags list -> unit
    corresponding closure will create a new reference, different from
    the global one.
 
-
    If [flags] contains [Marshal.Compat_32], marshaling fails when
-   it encounters an integer value outside the range [[-2{^30}, 2{^30}-1]]
+   it encounters an integer value outside the range [-2]{^[30]}, [2]{^[30]}[-1]
    of integers that are representable on a 32-bit platform.  This
    ensures that marshaled data generated on a 64-bit platform can be
    safely read back on a 32-bit platform.  If [flags] does not
    contain [Marshal.Compat_32], integer values outside the
-   range [[-2{^30}, 2{^30}-1]] are marshaled, and can be read back on
+   range [-2]{^[30]}, [2]{^[30]}[-1] are marshaled, and can be read back on
    a 64-bit platform, but will cause an error at un-marshaling time
    when read back on a 32-bit platform.  The [Mashal.Compat_32] flag
    only matters when marshaling is performed on a 64-bit platform;
    it has no effect if marshaling is performed on a 32-bit platform.
+   @raise Failure if [chan] is not in binary mode.
+
  *)
 
 external to_bytes :
-  'a -> extern_flags list -> bytes = "caml_output_value_to_string"
+  'a -> extern_flags list -> bytes = "caml_output_value_to_bytes"
 (** [Marshal.to_bytes v flags] returns a byte sequence containing
    the representation of [v].
    The [flags] argument has the same meaning as for
    {!Marshal.to_channel}.
-   @since 4.02.0 *)
+   @since 4.02 *)
 
 external to_string :
   'a -> extern_flags list -> string = "caml_output_value_to_string"
@@ -139,10 +141,10 @@ val from_channel : in_channel -> 'a
    one of the [Marshal.to_*] functions, and reconstructs and
    returns the corresponding value.
 
-   It raises [End_of_file] if the function has already reached the
-   end of file when starting to read from the channel, and raises
-   [Failure "input_value: truncated object"] if it reaches the end
-   of file later during the unmarshalling. *)
+   @raise End_of_file if [chan] is already at the end of the file.
+
+   @raise Failure if the end of the file is reached during
+   unmarshalling itself or if [chan] is not in binary mode.*)
 
 val from_bytes : bytes -> int -> 'a
 (** [Marshal.from_bytes buff ofs] unmarshals a structured value
@@ -150,7 +152,7 @@ val from_bytes : bytes -> int -> 'a
    representation is not read from a channel, but taken from
    the byte sequence [buff], starting at position [ofs].
    The byte sequence is not mutated.
-   @since 4.02.0 *)
+   @since 4.02 *)
 
 val from_string : string -> int -> 'a
 (** Same as [from_bytes] but take a string as argument instead of a
@@ -183,3 +185,12 @@ val data_size : bytes -> int -> int
 
 val total_size : bytes -> int -> int
 (** See {!Marshal.header_size}.*)
+
+(** {1:marshal_concurrency Marshal and domain safety}
+
+    Care must be taken when marshaling a mutable value that may be modified by
+    a different domain. Mutating a value that is being marshaled (i.e., turned
+    into a sequence of bytes) is a programming error and might result in
+    suprising values (when unmarshaling) due to tearing, since marshaling
+    involves byte-per-byte copy.
+*)
