@@ -1,18 +1,23 @@
 (* TEST
-* function_sections
-flags = "-S -function-sections"
-** arch_arm
-*** native
-reference = "${test_source_directory}/func_sections.arm.reference"
-** arch_arm64
-*** native
-reference = "${test_source_directory}/func_sections.arm.reference"
-** arch_amd64
-*** native
-reference = "${test_source_directory}/func_sections.reference"
-** arch_i386
-*** native
-reference = "${test_source_directory}/func_sections.reference"
+ flags = "-S -function-sections";
+ function_sections;
+ {
+   arch_arm;
+   reference = "${test_source_directory}/func_sections.arm.reference";
+   native;
+ }{
+   arch_arm64;
+   reference = "${test_source_directory}/func_sections.arm.reference";
+   native;
+ }{
+   arch_amd64;
+   reference = "${test_source_directory}/func_sections.reference";
+   native;
+ }{
+   arch_i386;
+   reference = "${test_source_directory}/func_sections.reference";
+   native;
+ }
 *)
 
 (* We have a separate reference output for ARM because
