@@ -1,7 +1,7 @@
 (* TEST
- flags = "-nostdlib -nopervasives -dlambda"
- * expect
- *)
+ flags = "-nostdlib -nopervasives -dlambda";
+ expect;
+*)
 
 (******************************************************************************)
 
@@ -18,10 +18,10 @@ let last_is_anys = function
   (last_is_anys/11 =
      (function param/13 : int
        (catch
-         (if (field 0 param/13) (if (field 1 param/13) (exit 1) 1)
-           (if (field 1 param/13) (exit 1) 2))
+         (if (field_imm 0 param/13) (if (field_imm 1 param/13) (exit 1) 1)
+           (if (field_imm 1 param/13) (exit 1) 2))
         with (1) 3)))
-  (apply (field 1 (global Toploop!)) "last_is_anys" last_is_anys/11))
+  (apply (field_mut 1 (global Toploop!)) "last_is_anys" last_is_anys/11))
 val last_is_anys : bool * bool -> int = <fun>
 |}]
 
@@ -35,10 +35,10 @@ let last_is_vars = function
   (last_is_vars/18 =
      (function param/22 : int
        (catch
-         (if (field 0 param/22) (if (field 1 param/22) (exit 3) 1)
-           (if (field 1 param/22) (exit 3) 2))
+         (if (field_imm 0 param/22) (if (field_imm 1 param/22) (exit 3) 1)
+           (if (field_imm 1 param/22) (exit 3) 2))
         with (3) 3)))
-  (apply (field 1 (global Toploop!)) "last_is_vars" last_is_vars/18))
+  (apply (field_mut 1 (global Toploop!)) "last_is_vars" last_is_vars/18))
 val last_is_vars : bool * bool -> int = <fun>
 |}]
 
@@ -55,9 +55,9 @@ type t = ..
   (A/26 = (makeblock 248 "A" (caml_fresh_oo_id 0))
    B/27 = (makeblock 248 "B" (caml_fresh_oo_id 0))
    C/28 = (makeblock 248 "C" (caml_fresh_oo_id 0)))
-  (seq (apply (field 1 (global Toploop!)) "A/26" A/26)
-    (apply (field 1 (global Toploop!)) "B/27" B/27)
-    (apply (field 1 (global Toploop!)) "C/28" C/28)))
+  (seq (apply (field_mut 1 (global Toploop!)) "A/26" A/26)
+    (apply (field_mut 1 (global Toploop!)) "B/27" B/27)
+    (apply (field_mut 1 (global Toploop!)) "C/28" C/28)))
 type t += A | B of unit | C of bool * int
 |}]
 
@@ -71,20 +71,20 @@ let f = function
 ;;
 [%%expect{|
 (let
-  (C/28 = (apply (field 0 (global Toploop!)) "C/28")
-   B/27 = (apply (field 0 (global Toploop!)) "B/27")
-   A/26 = (apply (field 0 (global Toploop!)) "A/26")
+  (C/28 = (apply (field_mut 0 (global Toploop!)) "C/28")
+   B/27 = (apply (field_mut 0 (global Toploop!)) "B/27")
+   A/26 = (apply (field_mut 0 (global Toploop!)) "A/26")
    f/29 =
      (function param/31 : int
-       (let (*match*/32 =a (field 0 param/31))
+       (let (*match*/32 =a (field_imm 0 param/31))
          (catch
-           (if (== *match*/32 A/26) (if (field 1 param/31) 1 (exit 8))
+           (if (== *match*/32 A/26) (if (field_imm 1 param/31) 1 (exit 8))
              (exit 8))
           with (8)
-           (if (field 1 param/31)
-             (if (== (field 0 *match*/32) B/27) 2
-               (if (== (field 0 *match*/32) C/28) 3 4))
-             (if (field 2 param/31) 12 11))))))
-  (apply (field 1 (global Toploop!)) "f" f/29))
+           (if (field_imm 1 param/31)
+             (if (== (field_imm 0 *match*/32) B/27) 2
+               (if (== (field_imm 0 *match*/32) C/28) 3 4))
+             (if (field_imm 2 param/31) 12 11))))))
+  (apply (field_mut 1 (global Toploop!)) "f" f/29))
 val f : t * bool * bool -> int = <fun>
 |}]
