@@ -1,5 +1,5 @@
 (* TEST
-   * expect
+ expect;
 *)
 
 type 'a x = private [> `x] as 'a;;
@@ -10,7 +10,7 @@ Error: Type declarations do not match:
        is not included in
          type 'a x
        Their parameters differ
-       The type 'b x as 'b is not equal to the type 'a
+       The type "'b x as 'b" is not equal to the type "'a"
 |}, Principal{|
 Line 1:
 Error: Type declarations do not match:
@@ -18,7 +18,7 @@ Error: Type declarations do not match:
        is not included in
          type 'a x
        Their parameters differ
-       The type [> `x ] is not equal to the type 'a
+       The type "[> `x ]" is not equal to the type "'a"
 |}];;
 
 
