@@ -1,5 +1,7 @@
 (* TEST
-   exit_status = "2"
+ exit_status = "2";
+ reason = "port stat-mem-prof : https://github.com/ocaml/ocaml/pull/8634";
+ skip;
 *)
 
 open Gc.Memprof
