@@ -1,5 +1,5 @@
 (* TEST
-   * expect
+ expect;
 *)
 
 type foo = ..
@@ -72,7 +72,7 @@ type bar += Bar of int
 Line 1, characters 0-22:
 1 | type bar += Bar of int
     ^^^^^^^^^^^^^^^^^^^^^^
-Error: Type definition bar is not extensible
+Error: Type definition "bar" is not extensible
 |}]
 
 type baz = bar = ..
@@ -81,8 +81,8 @@ type baz = bar = ..
 Line 1, characters 0-19:
 1 | type baz = bar = ..
     ^^^^^^^^^^^^^^^^^^^
-Error: This variant or record definition does not match that of type bar
-       Their kinds differ.
+Error: This variant or record definition does not match that of type "bar"
+       The original is abstract, but this is an extensible variant.
 |}]
 
 (* Abbreviations need to match parameters *)
@@ -99,7 +99,7 @@ type ('a, 'b) bar = 'a foo = ..
 Line 1, characters 0-31:
 1 | type ('a, 'b) bar = 'a foo = ..
     ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
-Error: This variant or record definition does not match that of type 'a foo
+Error: This variant or record definition does not match that of type "'a foo"
        They have different arities.
 |}]
 
@@ -116,8 +116,9 @@ Line 1, characters 0-37:
 1 | type ('a, 'b) bar = ('a, 'a) foo = ..
     ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
 Error: This variant or record definition does not match that of type
-         ('a, 'a) foo
-       Their constraints differ.
+         "('a, 'a) foo"
+       Their parameters differ
+       The type "'a" is not equal to the type "'b"
 |}]
 
 (* Check that signatures can hide exstensibility *)
@@ -146,7 +147,7 @@ type M_S.foo += Foo
 Line 1, characters 0-19:
 1 | type M_S.foo += Foo
     ^^^^^^^^^^^^^^^^^^^
-Error: Type definition M_S.foo is not extensible
+Error: Type definition "M_S.foo" is not extensible
 |}]
 
 (* Check that signatures cannot add extensibility *)
@@ -175,7 +176,7 @@ Error: Signature mismatch:
          type foo = M.foo
        is not included in
          type foo = ..
-       Their kinds differ.
+       The first is abstract, but the second is an extensible variant.
 |}]
 
 (* Check that signatures can make exstensibility private *)
@@ -236,7 +237,7 @@ Error: Signature mismatch:
          type foo = M.foo = private ..
        is not included in
          type foo = ..
-       A private type would be revealed.
+       A private extensible variant would be revealed.
 |}]
 
 
@@ -290,7 +291,7 @@ type exn += private Foobar
 Line 2, characters 14-20:
 2 | let _ = raise Foobar
                   ^^^^^^
-Error: Cannot use private constructor Foobar to create values of type exn
+Error: Cannot use private constructor "Foobar" to create values of type "exn"
 |}]
 
 
@@ -306,11 +307,12 @@ type foo += Foo
 Line 3, characters 8-26:
 3 | let f = function Foo -> ()
             ^^^^^^^^^^^^^^^^^^
-Warning 8: this pattern-matching is not exhaustive.
+Warning 8 [partial-match]: this pattern-matching is not exhaustive.
 Here is an example of a case that is not matched:
 *extension*
 Matching over values of extensible variant types (the *extension* above)
 must include a wild card pattern in order to be exhaustive.
+
 val f : foo -> unit = <fun>
 |}]
 
@@ -327,11 +329,12 @@ Lines 1-4, characters 8-11:
 2 |   | [Foo] -> 1
 3 |   | _::_::_ -> 3
 4 |   | [] -> 2
-Warning 8: this pattern-matching is not exhaustive.
+Warning 8 [partial-match]: this pattern-matching is not exhaustive.
 Here is an example of a case that is not matched:
 *extension*::[]
 Matching over values of extensible variant types (the *extension* above)
 must include a wild card pattern in order to be exhaustive.
+
 val f : foo list -> int = <fun>
 |}]
 
@@ -350,10 +353,11 @@ let f = function IPair (i, j) -> Format.sprintf "(%d, %d)" i j ;;
 Line 1, characters 8-62:
 1 | let f = function IPair (i, j) -> Format.sprintf "(%d, %d)" i j ;;
             ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
-Warning 8: this pattern-matching is not exhaustive.
+Warning 8 [partial-match]: this pattern-matching is not exhaustive.
 Here is an example of a case that is not matched:
 *extension*
 Matching over values of extensible variant types (the *extension* above)
 must include a wild card pattern in order to be exhaustive.
+
 val f : t -> string = <fun>
 |}]
