@@ -1,5 +1,5 @@
 (* TEST
-   * expect
+ expect;
 *)
 
 type _ t = C : ((('a -> 'o) -> 'o) -> ('b -> 'o) -> 'o) t
@@ -10,6 +10,6 @@ type _ t = C : ((('a -> 'o) -> 'o) -> ('b -> 'o) -> 'o) t
 Line 3, characters 24-25:
 3 |  fun C k -> k (fun x -> x);;
                             ^
-Error: This expression has type $0 but an expression was expected of type
-         $1 = ($2 -> $1) -> $1
+Error: This expression has type "$0" but an expression was expected of type
+         "$1" = "o"
 |}];;
