@@ -1,5 +1,5 @@
 (* TEST
-   * expect
+ expect;
 *)
 
 let f (type t) (x : t) = x
