@@ -1,5 +1,5 @@
 (* TEST
-   * expect
+ expect;
 *)
 
 module C = Char;;
@@ -18,13 +18,13 @@ module C' :
     external code : char -> int = "%identity"
     val chr : int -> char
     val escaped : char -> string
-    val lowercase : char -> char
-    val uppercase : char -> char
     val lowercase_ascii : char -> char
     val uppercase_ascii : char -> char
     type t = char
     val compare : t -> t -> int
     val equal : t -> t -> bool
+    val seeded_hash : int -> t -> int
+    val hash : t -> int
     external unsafe_chr : int -> char = "%identity"
   end
 - : char = 'B'
@@ -33,13 +33,13 @@ module C3 :
     external code : char -> int = "%identity"
     val chr : int -> char
     val escaped : char -> string
-    val lowercase : char -> char
-    val uppercase : char -> char
     val lowercase_ascii : char -> char
     val uppercase_ascii : char -> char
     type t = char
     val compare : t -> t -> int
     val equal : t -> t -> bool
+    val seeded_hash : int -> t -> int
+    val hash : t -> int
     external unsafe_chr : int -> char = "%identity"
   end
 - : char = 'B'
@@ -62,13 +62,13 @@ module F :
       external code : char -> int = "%identity"
       val chr : int -> char
       val escaped : char -> string
-      val lowercase : char -> char
-      val uppercase : char -> char
       val lowercase_ascii : char -> char
       val uppercase_ascii : char -> char
       type t = char
       val compare : t -> t -> int
       val equal : t -> t -> bool
+      val seeded_hash : int -> t -> int
+      val hash : t -> int
       external unsafe_chr : int -> char = "%identity"
     end
 module C4 :
@@ -76,13 +76,13 @@ module C4 :
     external code : char -> int = "%identity"
     val chr : int -> char
     val escaped : char -> string
-    val lowercase : char -> char
-    val uppercase : char -> char
     val lowercase_ascii : char -> char
     val uppercase_ascii : char -> char
     type t = char
     val compare : t -> t -> int
     val equal : t -> t -> bool
+    val seeded_hash : int -> t -> int
+    val hash : t -> int
     external unsafe_chr : int -> char = "%identity"
   end
 - : char = 'B'
@@ -280,8 +280,6 @@ module StringSet :
     type elt = String.t
     type t = Set.Make(String).t
     val empty : t
-    val is_empty : t -> bool
-    val mem : elt -> t -> bool
     val add : elt -> t -> t
     val singleton : elt -> t
     val remove : elt -> t -> t
@@ -289,17 +287,6 @@ module StringSet :
     val inter : t -> t -> t
     val disjoint : t -> t -> bool
     val diff : t -> t -> t
-    val compare : t -> t -> int
-    val equal : t -> t -> bool
-    val subset : t -> t -> bool
-    val iter : (elt -> unit) -> t -> unit
-    val map : (elt -> elt) -> t -> t
-    val fold : (elt -> 'a -> 'a) -> t -> 'a -> 'a
-    val for_all : (elt -> bool) -> t -> bool
-    val exists : (elt -> bool) -> t -> bool
-    val filter : (elt -> bool) -> t -> t
-    val filter_map : (elt -> elt option) -> t -> t
-    val partition : (elt -> bool) -> t -> t * t
     val cardinal : t -> int
     val elements : t -> elt list
     val min_elt : t -> elt
@@ -308,16 +295,31 @@ module StringSet :
     val max_elt_opt : t -> elt option
     val choose : t -> elt
     val choose_opt : t -> elt option
-    val split : elt -> t -> t * bool * t
     val find : elt -> t -> elt
     val find_opt : elt -> t -> elt option
     val find_first : (elt -> bool) -> t -> elt
     val find_first_opt : (elt -> bool) -> t -> elt option
     val find_last : (elt -> bool) -> t -> elt
     val find_last_opt : (elt -> bool) -> t -> elt option
+    val iter : (elt -> unit) -> t -> unit
+    val fold : (elt -> 'acc -> 'acc) -> t -> 'acc -> 'acc
+    val map : (elt -> elt) -> t -> t
+    val filter : (elt -> bool) -> t -> t
+    val filter_map : (elt -> elt option) -> t -> t
+    val partition : (elt -> bool) -> t -> t * t
+    val split : elt -> t -> t * bool * t
+    val is_empty : t -> bool
+    val mem : elt -> t -> bool
+    val equal : t -> t -> bool
+    val compare : t -> t -> int
+    val subset : t -> t -> bool
+    val for_all : (elt -> bool) -> t -> bool
+    val exists : (elt -> bool) -> t -> bool
+    val to_list : t -> elt list
     val of_list : elt list -> t
     val to_seq_from : elt -> t -> elt Seq.t
     val to_seq : t -> elt Seq.t
+    val to_rev_seq : t -> elt Seq.t
     val add_seq : elt Seq.t -> t -> t
     val of_seq : elt Seq.t -> t
   end
@@ -326,8 +328,6 @@ module SSet :
     type elt = S.t
     type t = Set.Make(S).t
     val empty : t
-    val is_empty : t -> bool
-    val mem : elt -> t -> bool
     val add : elt -> t -> t
     val singleton : elt -> t
     val remove : elt -> t -> t
@@ -335,17 +335,6 @@ module SSet :
     val inter : t -> t -> t
     val disjoint : t -> t -> bool
     val diff : t -> t -> t
-    val compare : t -> t -> int
-    val equal : t -> t -> bool
-    val subset : t -> t -> bool
-    val iter : (elt -> unit) -> t -> unit
-    val map : (elt -> elt) -> t -> t
-    val fold : (elt -> 'a -> 'a) -> t -> 'a -> 'a
-    val for_all : (elt -> bool) -> t -> bool
-    val exists : (elt -> bool) -> t -> bool
-    val filter : (elt -> bool) -> t -> t
-    val filter_map : (elt -> elt option) -> t -> t
-    val partition : (elt -> bool) -> t -> t * t
     val cardinal : t -> int
     val elements : t -> elt list
     val min_elt : t -> elt
@@ -354,16 +343,31 @@ module SSet :
     val max_elt_opt : t -> elt option
     val choose : t -> elt
     val choose_opt : t -> elt option
-    val split : elt -> t -> t * bool * t
     val find : elt -> t -> elt
     val find_opt : elt -> t -> elt option
     val find_first : (elt -> bool) -> t -> elt
     val find_first_opt : (elt -> bool) -> t -> elt option
     val find_last : (elt -> bool) -> t -> elt
     val find_last_opt : (elt -> bool) -> t -> elt option
+    val iter : (elt -> unit) -> t -> unit
+    val fold : (elt -> 'acc -> 'acc) -> t -> 'acc -> 'acc
+    val map : (elt -> elt) -> t -> t
+    val filter : (elt -> bool) -> t -> t
+    val filter_map : (elt -> elt option) -> t -> t
+    val partition : (elt -> bool) -> t -> t * t
+    val split : elt -> t -> t * bool * t
+    val is_empty : t -> bool
+    val mem : elt -> t -> bool
+    val equal : t -> t -> bool
+    val compare : t -> t -> int
+    val subset : t -> t -> bool
+    val for_all : (elt -> bool) -> t -> bool
+    val exists : (elt -> bool) -> t -> bool
+    val to_list : t -> elt list
     val of_list : elt list -> t
     val to_seq_from : elt -> t -> elt Seq.t
     val to_seq : t -> elt Seq.t
+    val to_rev_seq : t -> elt Seq.t
     val add_seq : elt Seq.t -> t -> t
     val of_seq : elt Seq.t -> t
   end
@@ -404,8 +408,6 @@ module A :
         type elt = B.t
         type t = Set.Make(B).t
         val empty : t
-        val is_empty : t -> bool
-        val mem : elt -> t -> bool
         val add : elt -> t -> t
         val singleton : elt -> t
         val remove : elt -> t -> t
@@ -413,17 +415,6 @@ module A :
         val inter : t -> t -> t
         val disjoint : t -> t -> bool
         val diff : t -> t -> t
-        val compare : t -> t -> int
-        val equal : t -> t -> bool
-        val subset : t -> t -> bool
-        val iter : (elt -> unit) -> t -> unit
-        val map : (elt -> elt) -> t -> t
-        val fold : (elt -> 'a -> 'a) -> t -> 'a -> 'a
-        val for_all : (elt -> bool) -> t -> bool
-        val exists : (elt -> bool) -> t -> bool
-        val filter : (elt -> bool) -> t -> t
-        val filter_map : (elt -> elt option) -> t -> t
-        val partition : (elt -> bool) -> t -> t * t
         val cardinal : t -> int
         val elements : t -> elt list
         val min_elt : t -> elt
@@ -432,16 +423,31 @@ module A :
         val max_elt_opt : t -> elt option
         val choose : t -> elt
         val choose_opt : t -> elt option
-        val split : elt -> t -> t * bool * t
         val find : elt -> t -> elt
         val find_opt : elt -> t -> elt option
         val find_first : (elt -> bool) -> t -> elt
         val find_first_opt : (elt -> bool) -> t -> elt option
         val find_last : (elt -> bool) -> t -> elt
         val find_last_opt : (elt -> bool) -> t -> elt option
+        val iter : (elt -> unit) -> t -> unit
+        val fold : (elt -> 'acc -> 'acc) -> t -> 'acc -> 'acc
+        val map : (elt -> elt) -> t -> t
+        val filter : (elt -> bool) -> t -> t
+        val filter_map : (elt -> elt option) -> t -> t
+        val partition : (elt -> bool) -> t -> t * t
+        val split : elt -> t -> t * bool * t
+        val is_empty : t -> bool
+        val mem : elt -> t -> bool
+        val equal : t -> t -> bool
+        val compare : t -> t -> int
+        val subset : t -> t -> bool
+        val for_all : (elt -> bool) -> t -> bool
+        val exists : (elt -> bool) -> t -> bool
+        val to_list : t -> elt list
         val of_list : elt list -> t
         val to_seq_from : elt -> t -> elt Seq.t
         val to_seq : t -> elt Seq.t
+        val to_rev_seq : t -> elt Seq.t
         val add_seq : elt Seq.t -> t -> t
         val of_seq : elt Seq.t -> t
       end
@@ -451,27 +457,44 @@ module A1 = A
 - : bool = true
 |}];;
 
-(* PR#3476 *)
-(* Does not work yet *)
+(* PR#3476: *)
 module FF(X : sig end) = struct type t end
 module M = struct
   module X = struct end
-  module Y = FF (X) (* XXX *)
+  module Y = FF (X)
   type t = Y.t
 end
 module F (Y : sig type t end) (M : sig type t = Y.t end) = struct end;;
 
 module G = F (M.Y);;
-(*module N = G (M);;
-module N = F (M.Y) (M);;*)
+module N = G (M);;
+module N = F (M.Y) (M);;
 [%%expect{|
 module FF : functor (X : sig end) -> sig type t end
 module M :
   sig module X : sig end module Y : sig type t = FF(X).t end type t = Y.t end
 module F : functor (Y : sig type t end) (M : sig type t = Y.t end) -> sig end
 module G : functor (M : sig type t = M.Y.t end) -> sig end
+module N : sig end
+module N : sig end
 |}];;
 
+(* PR#5058 *)
+module F (M : sig end) : sig type t end = struct type t = int end
+module T = struct
+module M = struct end
+include F(M)
+end
+include T
+let f (x : t) : T.t = x
+[%%expect {|
+module F : functor (M : sig end) -> sig type t end
+module T : sig module M : sig end type t = F(M).t end
+module M = T.M
+type t = F(M).t
+val f : t -> T.t = <fun>
+|}]
+
 (* PR#6307 *)
 
 module A1 = struct end
@@ -517,8 +540,6 @@ module SInt :
     type elt = Int.t
     type t = Set.Make(Int).t
     val empty : t
-    val is_empty : t -> bool
-    val mem : elt -> t -> bool
     val add : elt -> t -> t
     val singleton : elt -> t
     val remove : elt -> t -> t
@@ -526,17 +547,6 @@ module SInt :
     val inter : t -> t -> t
     val disjoint : t -> t -> bool
     val diff : t -> t -> t
-    val compare : t -> t -> int
-    val equal : t -> t -> bool
-    val subset : t -> t -> bool
-    val iter : (elt -> unit) -> t -> unit
-    val map : (elt -> elt) -> t -> t
-    val fold : (elt -> 'a -> 'a) -> t -> 'a -> 'a
-    val for_all : (elt -> bool) -> t -> bool
-    val exists : (elt -> bool) -> t -> bool
-    val filter : (elt -> bool) -> t -> t
-    val filter_map : (elt -> elt option) -> t -> t
-    val partition : (elt -> bool) -> t -> t * t
     val cardinal : t -> int
     val elements : t -> elt list
     val min_elt : t -> elt
@@ -545,16 +555,31 @@ module SInt :
     val max_elt_opt : t -> elt option
     val choose : t -> elt
     val choose_opt : t -> elt option
-    val split : elt -> t -> t * bool * t
     val find : elt -> t -> elt
     val find_opt : elt -> t -> elt option
     val find_first : (elt -> bool) -> t -> elt
     val find_first_opt : (elt -> bool) -> t -> elt option
     val find_last : (elt -> bool) -> t -> elt
     val find_last_opt : (elt -> bool) -> t -> elt option
+    val iter : (elt -> unit) -> t -> unit
+    val fold : (elt -> 'acc -> 'acc) -> t -> 'acc -> 'acc
+    val map : (elt -> elt) -> t -> t
+    val filter : (elt -> bool) -> t -> t
+    val filter_map : (elt -> elt option) -> t -> t
+    val partition : (elt -> bool) -> t -> t * t
+    val split : elt -> t -> t * bool * t
+    val is_empty : t -> bool
+    val mem : elt -> t -> bool
+    val equal : t -> t -> bool
+    val compare : t -> t -> int
+    val subset : t -> t -> bool
+    val for_all : (elt -> bool) -> t -> bool
+    val exists : (elt -> bool) -> t -> bool
+    val to_list : t -> elt list
     val of_list : elt list -> t
     val to_seq_from : elt -> t -> elt Seq.t
     val to_seq : t -> elt Seq.t
+    val to_rev_seq : t -> elt Seq.t
     val add_seq : elt Seq.t -> t -> t
     val of_seq : elt Seq.t -> t
   end
@@ -574,7 +599,7 @@ module Int2 : sig type t = int val compare : 'a -> 'a -> int end
 Line 15, characters 10-30:
 15 |   include S with module I := I
                ^^^^^^^^^^^^^^^^^^^^
-Error: In this `with' constraint, the new definition of I
+Error: In this "with" constraint, the new definition of "I"
        does not match its original definition in the constrained signature:
        Modules do not match: (module Int2) is not included in (module Int)
 |}];;
@@ -669,13 +694,13 @@ Error: Module type declarations do not match:
          module type A = sig module M = F(List) end
        does not match
          module type A = sig module M = F(List) end
-       At position module type A = <here>
-       Modules do not match:
+       At position "module type A = <here>"
+       Module types do not match:
          sig module M = F(List) end
-       is not included in
+       is not equal to
          sig module M = F(List) end
-       At position module type A = sig module M : <here> end
-       Module F(List) cannot be aliased
+       At position "module type A = sig module M : <here> end"
+       Module "F(List)" cannot be aliased
 |}];;
 
 (* Shinwell 2014-04-23 *)
