@@ -1,5 +1,5 @@
 (* TEST
-   * expect
+ expect;
 *)
 
 (* Using generative functors *)
@@ -42,27 +42,35 @@ module H : functor () -> S
 
 (* Alias *)
 module U = struct end;;
-module M = F(struct end);; (* ok *)
+module M1 = F();; (* ok *)
+module M2 = F(struct end);; (* accepted with a warning *)
 [%%expect{|
 module U : sig end
-module M : S
+module M1 : S
+Line 3, characters 14-24:
+3 | module M2 = F(struct end);; (* accepted with a warning *)
+                  ^^^^^^^^^^
+Warning 73 [generative-application-expects-unit]: A generative functor
+should be applied to '()'; using '(struct end)' is deprecated.
+
+module M2 : S
 |}];;
 module M = F(U);; (* fail *)
 [%%expect{|
 Line 1, characters 11-12:
 1 | module M = F(U);; (* fail *)
                ^
-Error: This is a generative functor. It can only be applied to ()
+Error: This is a generative functor. It can only be applied to "()"
 |}];;
 
 (* Cannot coerce between applicative and generative *)
 module F1 (X : sig end) = struct end;;
-module F2 : functor () -> sig end = F1;; (* fail *)
+module F2 : () -> sig end = F1;; (* fail *)
 [%%expect{|
 module F1 : functor (X : sig end) -> sig end
-Line 2, characters 36-38:
-2 | module F2 : functor () -> sig end = F1;; (* fail *)
-                                        ^^
+Line 2, characters 28-30:
+2 | module F2 : () -> sig end = F1;; (* fail *)
+                                ^^
 Error: Signature mismatch:
        Modules do not match:
          functor (X : sig end) -> ...
@@ -98,3 +106,12 @@ module Y : functor (X : sig end) (Y : sig end) (Z : sig end) -> sig end
 module Z : sig end -> sig end -> sig end -> sig end
 module GZ : functor (X : sig end) () (Z : sig end) -> sig end
 |}];;
+
+(* disabling warning 73 in the argument *)
+module F5 () = struct end;;
+module No_warn = F5 (struct end [@warning "-73"])
+
+[%%expect{|
+module F5 : functor () -> sig end
+module No_warn : sig end
+|}]
