@@ -1,5 +1,5 @@
 (* TEST
-   * expect
+ expect;
 *)
 
 module type S = sig
@@ -12,7 +12,7 @@ module Good (X : S with type t := unit) = struct
 end;;
 [%%expect{|
 module type S = sig type t val x : t end
-module Good : functor (X : sig val x : unit end) -> sig  end
+module Good : functor (X : sig val x : unit end) -> sig end
 |}];;
 
 module type T = sig module M : S end;;
@@ -23,6 +23,5 @@ end;;
 [%%expect{|
 module type T = sig module M : S end
 module Bad :
-  functor (X : sig module M : sig type t = unit val x : t end end) ->
-    sig  end
+  functor (X : sig module M : sig type t = unit val x : t end end) -> sig end
 |}];;
