@@ -1,5 +1,5 @@
 (* TEST
-   * expect
+ expect;
 *)
 
 module F (X : sig type t = private < foo:int; ..> val x : t end) = struct
@@ -37,5 +37,5 @@ module A : sig end = struct
   let _ = (N.x = M.x)
 end;;
 [%%expect{|
-module A : sig  end
+module A : sig end
 |}]
