@@ -1,5 +1,5 @@
 (* TEST
-   * expect
+ expect;
 *)
 
 (* Subtyping is "syntactic" *)
@@ -37,7 +37,7 @@ Lines 3-5, characters 4-3:
 5 | end..
 Error: Some type variables are unbound in this type:
          class d : unit -> object method f : 'a -> unit end
-       The method f has type 'a -> unit where 'a is unbound
+       The method "f" has type "'a -> unit" where "'a" is unbound
 |}];;
 
 (* Create instance #c *)
@@ -108,7 +108,7 @@ Lines 1-3, characters 13-3:
 2 |   method virtual f : int
 3 | end..
 Error: This non-virtual class has virtual methods.
-       The following methods are virtual : f
+       The following methods are virtual : "f"
 |}];;
 (* The class x should be virtual:  its methods f is undefined *)
 
@@ -123,7 +123,7 @@ Line 1, characters 49-57:
 1 | class virtual c ((x : 'a): < f : int >) = object (_ : 'a) end
                                                      ^^^^^^^^
 Error: This pattern cannot match self: it only matches values of type
-       < f : int >
+       "< f : int >"
 |}];;
 
 (* Constraint not respected *)
@@ -137,8 +137,8 @@ Lines 1-4, characters 0-3:
 2 |   constraint 'a = int
 3 |   method f x = (x : bool c)
 4 | end..
-Error: The abbreviation c is used with parameters bool c
-       which are incompatible with constraints int c
+Error: The abbreviation "c" is used with parameter(s) "bool "
+       which are incompatible with constraint(s) "int "
 |}];;
 
 (* Different constraints *)
@@ -183,9 +183,10 @@ Lines 1-3, characters 0-3:
 2 |   method f = (x : 'a)
 3 | end..
 Error: The type of this class,
-       class ['a] c :
-         unit -> object constraint 'a = '_weak1 list ref method f : 'a end,
-       contains type variables that cannot be generalized
+       "class ['a] c :
+         unit -> object constraint 'a = '_weak1 list ref method f : 'a end",
+       contains the non-generalizable type variable(s): "'_weak1".
+       (see manual section 6.1.2)
 |}];;
 
 (* Abbreviations *)
@@ -196,12 +197,14 @@ Line 1, characters 0-32:
 1 | type 'a c = <f : 'a c; g : 'a d>
     ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
 Error: This recursive type is not regular.
-       The type constructor c is defined as
-         type 'a c
+       The type constructor "c" is defined as
+         type "'a c"
        but it is used as
-         int c
+         "int c"
        after the following expansion(s):
-         'a d = < f : int c >
+         "< f : 'a c; g : 'a d >" contains "'a d",
+         "'a d" = "< f : int c >",
+         "< f : int c >" contains "int c"
        All uses need to match the definition for the recursive type to be regular.
 |}];;
 type 'a c = <f : 'a c; g : 'a d>
@@ -222,8 +225,10 @@ and 'a t = 'a t u;;
 Line 2, characters 0-17:
 2 | and 'a t = 'a t u;;
     ^^^^^^^^^^^^^^^^^
-Error: The definition of t contains a cycle:
-       'a t u
+Error: The type abbreviation "t" is cyclic:
+         "'a t u" contains "'a t",
+         "'a t" = "'a t u",
+         "'a t u" contains "'a t"
 |}];; (* fails since 4.04 *)
 type 'a u = 'a
 and 'a t = 'a t u;;
@@ -231,7 +236,9 @@ and 'a t = 'a t u;;
 Line 2, characters 0-17:
 2 | and 'a t = 'a t u;;
     ^^^^^^^^^^^^^^^^^
-Error: The type abbreviation t is cyclic
+Error: The type abbreviation "t" is cyclic:
+         "'a t" = "'a t u",
+         "'a t u" = "'a t"
 |}];;
 type 'a u = 'a;;
 [%%expect{|
@@ -242,7 +249,10 @@ type t = t u * t u;;
 Line 1, characters 0-18:
 1 | type t = t u * t u;;
     ^^^^^^^^^^^^^^^^^^
-Error: The type abbreviation t is cyclic
+Error: The type abbreviation "t" is cyclic:
+         "t" = "t u * t u",
+         "t u * t u" contains "t u",
+         "t u" = "t"
 |}];;
 
 type t = <x : 'a> as 'a;;
@@ -406,8 +416,8 @@ class c () = object val x = - true val y = -. () end;;
 Line 1, characters 30-34:
 1 | class c () = object val x = - true val y = -. () end;;
                                   ^^^^
-Error: This expression has type bool but an expression was expected of type
-         int
+Error: This expression has type "bool" but an expression was expected of type
+         "int"
 |}];;
 
 class c () = object method f = 1 method g = 1 method h = 1 end;;
@@ -475,19 +485,23 @@ Line 3, characters 2-13:
       ^^^^^^^^^^^
 Warning 13 [instance-variable-override]: the following instance variables are overridden by the class c :
   x
+
 Line 4, characters 6-7:
 4 |   val y = 3
           ^
 Warning 13 [instance-variable-override]: the instance variable y is overridden.
+
 Line 6, characters 2-13:
 6 |   inherit d 7
       ^^^^^^^^^^^
 Warning 13 [instance-variable-override]: the following instance variables are overridden by the class d :
   t z
+
 Line 7, characters 6-7:
 7 |   val u = 3
           ^
 Warning 13 [instance-variable-override]: the instance variable u is overridden.
+
 class e :
   unit ->
   object
@@ -642,8 +656,8 @@ Lines 1-4, characters 0-3:
 2 |   val m = Array.make_matrix sz sz init
 3 |   method add (mtx : 'a matrix) = (mtx#m.(0).(0) : 'a)
 4 | end..
-Error: The abbreviation 'a matrix expands to type < add : 'a matrix -> 'a >
-       but is used with type < m : 'a array array; .. >
+Error: The abbreviation "'a matrix" expands to type "< add : 'a matrix -> 'a >"
+       but is used with type "< m : 'a array array; .. >"
 |}];;
 
 class c () = object method m = new c () end;;
@@ -699,10 +713,10 @@ Error: Signature mismatch:
          val f : (#c as 'a) -> 'a
        is not included in
          val f : #c -> #c
-       The type (#c as 'a) -> 'a is not compatible with the type #c -> #c
-       Type #c as 'a = < m : 'a; .. > is not compatible with type
-         #c as 'b = < m : 'b; .. >
-       Type 'a is not compatible with type 'b
+       The type "(#c as 'a) -> 'a" is not compatible with the type "#c -> #c"
+       Type "#c as 'a" = "< m : 'a; .. >" is not compatible with type
+         "#c as 'b" = "< m : 'b; .. >"
+       Type "'a" is not compatible with type "'b"
 |}];;
 
 module M = struct type t = int class t () = object end end;;
@@ -710,7 +724,7 @@ module M = struct type t = int class t () = object end end;;
 Line 1, characters 37-38:
 1 | module M = struct type t = int class t () = object end end;;
                                          ^
-Error: Multiple definition of the type name t.
+Error: Multiple definition of the type name "t".
        Names must be unique in a given structure or signature.
 |}];;
 
@@ -724,16 +738,16 @@ fun x -> (x : int -> bool :> 'a -> 'a);;
 Line 1, characters 9-38:
 1 | fun x -> (x : int -> bool :> 'a -> 'a);;
              ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
-Error: Type int -> bool is not a subtype of int -> int
-       Type bool is not a subtype of int
+Error: Type "int -> bool" is not a subtype of "int -> int"
+       Type "bool" is not a subtype of "int"
 |}];;
 fun x -> (x : int -> bool :> int -> int);;
 [%%expect{|
 Line 1, characters 9-40:
 1 | fun x -> (x : int -> bool :> int -> int);;
              ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
-Error: Type int -> bool is not a subtype of int -> int
-       Type bool is not a subtype of int
+Error: Type "int -> bool" is not a subtype of "int -> int"
+       Type "bool" is not a subtype of "int"
 |}];;
 fun x -> (x : < > :> < .. >);;
 [%%expect{|
@@ -764,20 +778,36 @@ type 'a t
 |}];;
 fun (x : 'a t as 'a) -> ();;
 [%%expect{|
-Line 1, characters 9-19:
+Line 1, characters 17-19:
 1 | fun (x : 'a t as 'a) -> ();;
-             ^^^^^^^^^^
-Error: This alias is bound to type 'a t but is used as an instance of type 'a
-       The type variable 'a occurs inside 'a t
+                     ^^
+Error: This alias is bound to type "'a t" but is used as an instance of type "'a"
+       The type variable "'a" occurs inside "'a t"
 |}];;
 fun (x : 'a t) -> (x : 'a); ();;
 [%%expect{|
 Line 1, characters 19-20:
 1 | fun (x : 'a t) -> (x : 'a); ();;
                        ^
-Error: This expression has type 'a t but an expression was expected of type
-         'a
-       The type variable 'a occurs inside 'a t
+Error: This expression has type "'a t" but an expression was expected of type
+         "'a"
+       The type variable "'a" occurs inside "'a t"
+|}];;
+fun ((x : 'a) | (x : 'a t)) -> ();;
+[%%expect{|
+Line 1, characters 10-12:
+1 | fun ((x : 'a) | (x : 'a t)) -> ();;
+              ^^
+Error: This type "'a t" should be an instance of type "'a"
+       The type variable "'a" occurs inside "'a t"
+|}];;
+fun ((x : 'a) | (x : 'a t)) -> ();;
+[%%expect{|
+Line 1, characters 10-12:
+1 | fun ((x : 'a) | (x : 'a t)) -> ();;
+              ^^
+Error: This type "'a t" should be an instance of type "'a"
+       The type variable "'a" occurs inside "'a t"
 |}];;
 type 'a t = < x : 'a >;;
 [%%expect{|
@@ -793,8 +823,18 @@ Line 1, characters 18-26:
 1 | fun (x : 'a t) -> (x : 'a); ();;
                       ^^^^^^^^
 Warning 10 [non-unit-statement]: this expression should have type unit.
+
 - : ('a t as 'a) t -> unit = <fun>
 |}];;
+fun ((x : 'a) | (x : 'a t)) -> ();;
+[%%expect{|
+Line 1, characters 17-18:
+1 | fun ((x : 'a) | (x : 'a t)) -> ();;
+                     ^
+Warning 12 [redundant-subpat]: this sub-pattern is unused.
+
+- : ('a t as 'a) -> unit = <fun>
+|}];;
 
 class ['a] c () = object
   constraint 'a = < .. > -> unit
@@ -896,7 +936,7 @@ class a = object val x = 3 val y = x + 2 end;;
 Line 1, characters 35-36:
 1 | class a = object val x = 3 val y = x + 2 end;;
                                        ^
-Error: The instance variable x
+Error: The instance variable "x"
        cannot be accessed from the definition of another instance variable
 |}];;
 
@@ -905,7 +945,7 @@ class a = object (self) val x = self#m method m = 3 end;;
 Line 1, characters 32-36:
 1 | class a = object (self) val x = self#m method m = 3 end;;
                                     ^^^^
-Error: The self variable self
+Error: The self variable "self"
        cannot be accessed from the definition of an instance variable
 |}];;
 
@@ -916,7 +956,7 @@ class a : object method m : int end
 Line 2, characters 44-49:
 2 | class b = object inherit a as super val x = super#m end;;
                                                 ^^^^^
-Error: The ancestor variable super
+Error: The ancestor variable "super"
        cannot be accessed from the definition of an instance variable
 |}];;
 
@@ -936,7 +976,7 @@ end;;
 Line 2, characters 13-58:
 2 |   method o = object(_ : 'self) method o = assert false end
                  ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
-Error: Cannot close type of object literal: < o : '_weak3; _.. >
+Error: Cannot close type of object literal: "< o : '_weak4; .. > as '_weak3"
        it has been unified with the self type of a class that is not yet
        completely defined.
 |}];;
@@ -952,6 +992,7 @@ Line 4, characters 17-23:
 4 |       method n = self#m
                      ^^^^^^
 Warning 17 [undeclared-virtual-method]: the virtual method m is not declared.
+
 class c : object method m : int method n : int end
 |}];;
 
@@ -980,8 +1021,8 @@ class ['a] c : object ('a) constraint 'a = < .. > end
 Line 4, characters 14-25:
 4 |     inherit [ < m : int > ] c
                   ^^^^^^^^^^^
-Error: The type parameter < m : int >
-       does not meet its constraint: it should be < .. >
+Error: The type parameter "< m : int >"
+       does not meet its constraint: it should be "< .. >"
        Self type cannot be unified with a closed object type
 |}];;
 
@@ -997,8 +1038,8 @@ Error: The class type object method a : int method b : 'a end
        The class type object method a : int method b : 'a end
        is not matched by the class type
          object method a : 'a method b : 'a end
-       The method a has type int but is expected to have type 'a
-       Type int is not compatible with type 'a
+       The method a has type "int" but is expected to have type "'a"
+       Type "int" is not compatible with type "'a"
 |}];;
 
 class type ['a] ct = object ('a) end
@@ -1009,7 +1050,7 @@ Line 2, characters 10-31:
 2 | class c : [ < a : int; ..> ] ct = object method a = 3 end;;
               ^^^^^^^^^^^^^^^^^^^^^
 Error: This non-virtual class has undeclared virtual methods.
-       The following methods were not declared : a
+       The following methods were not declared : "a"
 |}];;
 
 class virtual c : [ < a : int; ..> ] ct = object method a = 3 end;;
@@ -1030,9 +1071,9 @@ Lines 3-5, characters 8-3:
 Error: The class type object ('a) method m : < m : 'a; .. > as 'a end
        is not matched by the class type
          object method m : < m : 'a > as 'a end
-       The method m has type < m : 'a; .. > as 'a
-       but is expected to have type < m : 'b > as 'b
-       Type 'a is not compatible with type <  >
+       The method m has type "< m : 'a; .. > as 'a"
+       but is expected to have type "< m : 'b > as 'b"
+       Type "'a" is not compatible with type "<  >"
 |}];;
 
 class c :
@@ -1051,10 +1092,10 @@ Error: The class type
          object method foo : (< foo : int; .. > as 'a) -> 'a -> unit end
        is not matched by the class type
          object method foo : < foo : int; .. > -> < foo : int > -> unit end
-       The method foo has type 'a. (< foo : int; .. > as 'a) -> 'a -> unit
+       The method foo has type "'a. (< foo : int; .. > as 'a) -> 'a -> unit"
        but is expected to have type
-         'b. (< foo : int; .. > as 'b) -> < foo : int > -> unit
-       Type 'c is not compatible with type <  >
+         "'b. (< foo : int; .. > as 'b) -> < foo : int > -> unit"
+       Type "'c" is not compatible with type "<  >"
 |}];;
 
 
@@ -1089,10 +1130,10 @@ Error: The class type
          object ('a) method m : (< m : 'a -> unit; .. > as 'a) -> unit end
        is not matched by the class type
          object method m : < m : 'a; x : int; .. > -> unit as 'a end
-       The method m has type (< m : 'a -> unit; .. > as 'a) -> unit
+       The method m has type "(< m : 'a -> unit; .. > as 'a) -> unit"
        but is expected to have type
-         'b. (< m : 'c; x : int; .. > as 'b) -> unit as 'c
-       Type 'a is not compatible with type < x : int; .. >
+         "'b. (< m : 'c; x : int; .. > as 'b) -> unit as 'c"
+       Type "'a" is not compatible with type "< x : int; .. >"
 |}];;
 
 let is_empty (x : < >) = ()
@@ -1102,8 +1143,8 @@ val is_empty : <  > -> unit = <fun>
 Line 2, characters 54-58:
 2 | class c = object (self) method private foo = is_empty self end;;
                                                           ^^^^
-Error: This expression has type < .. > but an expression was expected of type
-         <  >
+Error: This expression has type "< .. >" but an expression was expected of type
+         "<  >"
        Self type cannot be unified with a closed object type
 |}];;
 
@@ -1118,6 +1159,7 @@ Line 3, characters 10-75:
               ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
 Warning 15 [implicit-public-methods]: the following private methods were made public implicitly:
  foo.
+
 class c : object method foo : int end
 |}];;
 
@@ -1142,7 +1184,7 @@ Line 1, characters 10-45:
 1 | class c = object method virtual foo : int end;;
               ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
 Error: This non-virtual class has virtual methods.
-       The following methods are virtual : foo
+       The following methods are virtual : "foo"
 |}];;
 
 class type ct = object method virtual foo : int end;;
@@ -1151,7 +1193,7 @@ Line 1, characters 16-51:
 1 | class type ct = object method virtual foo : int end;;
                     ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
 Error: This non-virtual class type has virtual methods.
-       The following methods are virtual : foo
+       The following methods are virtual : "foo"
 |}];;
 
 let o = object method virtual foo : int end;;
@@ -1160,7 +1202,7 @@ Line 1, characters 8-43:
 1 | let o = object method virtual foo : int end;;
             ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
 Error: This object has virtual methods.
-       The following methods are virtual : foo
+       The following methods are virtual : "foo"
 |}];;
 
 class c = object(self) initializer self#foo end;;
@@ -1168,7 +1210,7 @@ class c = object(self) initializer self#foo end;;
 Line 1, characters 35-39:
 1 | class c = object(self) initializer self#foo end;;
                                        ^^^^
-Error: This expression has no method foo
+Error: This expression has no method "foo"
 |}];;
 
 let o = object(self) initializer self#foo end;;
@@ -1176,7 +1218,7 @@ let o = object(self) initializer self#foo end;;
 Line 1, characters 33-37:
 1 | let o = object(self) initializer self#foo end;;
                                      ^^^^
-Error: This expression has no method foo
+Error: This expression has no method "foo"
 |}];;
 
 let has_foo (x : < foo : int; ..>) = ()
@@ -1187,7 +1229,7 @@ Line 2, characters 10-51:
 2 | class c = object(self) initializer has_foo self end;;
               ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
 Error: This non-virtual class has undeclared virtual methods.
-       The following methods were not declared : foo
+       The following methods were not declared : "foo"
 |}];;
 
 let o = object(self) initializer has_foo self end;;
@@ -1195,9 +1237,9 @@ let o = object(self) initializer has_foo self end;;
 Line 1, characters 41-45:
 1 | let o = object(self) initializer has_foo self end;;
                                              ^^^^
-Error: This expression has type <  > but an expression was expected of type
-         < foo : int; .. >
-       The first object type has no method foo
+Error: This expression has type "<  >" but an expression was expected of type
+         "< foo : int; .. >"
+       The first object type has no method "foo"
 |}];;
 
 class c = object(_ : < foo : int; ..>) end;;
@@ -1206,7 +1248,7 @@ Line 1, characters 10-42:
 1 | class c = object(_ : < foo : int; ..>) end;;
               ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
 Error: This non-virtual class has undeclared virtual methods.
-       The following methods were not declared : foo
+       The following methods were not declared : "foo"
 |}];;
 
 class type ct = object(< foo : int; ..>) end;;
@@ -1215,7 +1257,7 @@ Line 1, characters 16-44:
 1 | class type ct = object(< foo : int; ..>) end;;
                     ^^^^^^^^^^^^^^^^^^^^^^^^^^^^
 Error: This non-virtual class type has undeclared virtual methods.
-       The following methods were not declared : foo
+       The following methods were not declared : "foo"
 |}];;
 
 let o = object(_ : < foo : int; ..>) end;;
@@ -1224,7 +1266,7 @@ Line 1, characters 8-40:
 1 | let o = object(_ : < foo : int; ..>) end;;
             ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
 Error: This object has undeclared virtual methods.
-       The following methods were not declared : foo
+       The following methods were not declared : "foo"
 |}];;
 
 (* Shadowing/overriding methods in class types *)
@@ -1288,8 +1330,8 @@ class c : object method private test : unit end
 Line 6, characters 9-16:
 6 | let () = (new c)#test
              ^^^^^^^
-Error: This expression has type c
-       It has no method test
+Error: This expression has type "c"
+       It has no method "test"
 |}];;
 
 class c = object
@@ -1318,8 +1360,8 @@ class c : object method private test : unit end
 Line 10, characters 9-16:
 10 | let () = (new c)#test
               ^^^^^^^
-Error: This expression has type c
-       It has no method test
+Error: This expression has type "c"
+       It has no method "test"
 |}];;
 
 class c = object
@@ -1357,7 +1399,7 @@ Lines 1-3, characters 10-3:
 2 |     method virtual m : int
 3 | end.........
 Error: This non-virtual class type has virtual methods.
-       The following methods are virtual : m
+       The following methods are virtual : "m"
 |}];;
 
 class virtual c : object
@@ -1382,7 +1424,7 @@ Line 5, characters 10-12:
 5 | class c : cv = object
               ^^
 Error: This non-virtual class type has virtual methods.
-       The following methods are virtual : m
+       The following methods are virtual : "m"
 |}];;
 
 class virtual c : cv = object
@@ -1398,7 +1440,7 @@ Line 1, characters 10-12:
 1 | class c = cv
               ^^
 Error: This non-virtual class has virtual methods.
-       The following methods are virtual : m
+       The following methods are virtual : "m"
 |}];;
 
 class virtual c = cv
@@ -1415,6 +1457,6 @@ Line 2, characters 0-16:
 2 | class d = ['a] c
     ^^^^^^^^^^^^^^^^
 Error: Some type variables are unbound in this type: class d : ['a] c
-       The method m has type (< f : 'b; x : 'a > as 'b) -> unit where 'a
+       The method "m" has type "(< f : 'b; x : 'a > as 'b) -> unit" where "'a"
        is unbound
 |}]
