@@ -1,8 +1,8 @@
 (* TEST
-include config
-include testing
-binary_modules = "config build_path_prefix_map misc identifiable numbers"
-* bytecode
+ include config;
+ include testing;
+ binary_modules = "config build_path_prefix_map misc identifiable numbers";
+ bytecode;
 *)
 
 let print_int i =
